# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Self

import torch

from ..metric import Metric
from ..utils import broadcast_to_prefix


class Mean(torch.nn.Module, Metric):
    """A class tracking the (optionally weighted) mean of given values."""

    def __init__(self, device: torch.device | None = None) -> None:
        """Create the Mean metric object."""
        super().__init__()
        self.register_buffer("_total", torch.tensor(0.0, dtype=torch.float32, device=device), persistent=False)
        self.register_buffer("_count", torch.tensor(0.0, dtype=torch.float32, device=device), persistent=False)

    @torch.no_grad
    def update(
        self, y: torch.Tensor, y_true: torch.Tensor | None = None, sample_weights: torch.Tensor | None = None,
    ) -> Self:
        """Update the accumulated mean by introducing new values.

        Optional sample weight might be provided; if not, all values are weighted with 1.

        Parameters:
          y: The values to average.
          y_true: This parameter is present for [npfl138.Metric][] compatibility, but must be `None` for this metric.
          sample_weights: Optional sample weights. Their shape must be broadcastable to a prefix of the shape of `y`.

        Returns:
          self
        """
        assert y_true is None, "The y_true parameter must be None for the Mean metric."

        if sample_weights is not None:
            sample_weights = broadcast_to_prefix(sample_weights, y.shape)

        self._total.add_(torch.sum(y * sample_weights) if sample_weights is not None else torch.sum(y))
        self._count.add_(torch.sum(sample_weights) if sample_weights is not None else y.numel())

        return self

    def compute(self) -> torch.Tensor:
        return self._total / self._count

    def reset(self) -> Self:
        self._total.zero_()
        self._count.zero_()
        return self



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Protocol

import torch


class Loss(Protocol):
    """An abstract loss function interface."""

    def __call__(self, y: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
        """Compute loss of the given predictions and gold targets.

        Parameters:
          y: The predicted outputs.
          y_true: The ground-truth targets.

        Returns:
          A tensor representing the computed loss.
        """
        ...



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Self

import torch

from .mean import Mean
from ..utils import maybe_remove_one_singleton_dimension


class BinaryAccuracy(Mean):
    """Binary classification accuracy metric.

    The predictions are assumed to be logits or probabilities predicted by a model,
    while the ground-truth targets are binary (0 or 1) values. In both cases, the predicted
    class is considered to be the one with larger probability.
    """

    def __init__(self, *, probs: bool = False, device: torch.device | None = None) -> None:
        """Create the BinaryAccuracy metric object.

        Parameters:
          probs: If `False`, the predictions are assumed to be logits; if `True`, the
            predictions are assumed to be probabilities. Note that gold targets are
            always expected to be probabilities.
        """
        super().__init__(device)
        self._probs = probs

    @torch.no_grad
    def update(
        self, y: torch.Tensor, y_true: torch.Tensor, sample_weights: torch.Tensor | None = None,
    ) -> Self:
        """Update the accumulated binary accuracy using new predictions and gold labels.

        Optional sample weight might be provided; if not, all values are weighted with 1.

        Parameters:
          y: The predicted outputs. Their shape either has to be exactly the same as `y_true` (no broadcasting),
            or can contain an additional single dimension of size 1. We consider the more probable class
            to be predicted.
          y_true: The ground-truth targets; they are rounded to 0 or 1 to obtain binary labels.
          sample_weights: Optional sample weights. If provided, their shape must be broadcastable
            to a prefix of a shape of `y_true`, and the loss for each sample is weighted accordingly.

        Returns:
          self
        """
        y = maybe_remove_one_singleton_dimension(y, y_true)
        assert y.shape == y_true.shape, f"Shapes of y {y.shape} and y_true {y_true.shape} have to match " \
            "up to one singleton dim in y."

        y = (y > (0.5 if self._probs else 0.0))
        y_true = (y_true > 0.5)

        return super().update(y == y_true, sample_weights=sample_weights)



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

seen_tags: set[str] = set()


def first_time(tag: str) -> bool:
    """Returns `True` when first called with the given `tag`."""
    if tag in seen_tags:
        return False

    seen_tags.add(tag)
    return True



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import os
from typing import Any, Self

import torch
import torch.utils.tensorboard

from .base_logger import BaseLogger
from ..type_aliases import AnyArray, DataFormat, TensorOrTensors


class TensorBoardLogger(BaseLogger):
    """A TensorBoard logger interface.

    In addition to implementing the [Logger][npfl138.Logger] interface, it provides
    a method for obtaining the underlying TensorBoard
    [SummaryWriter][torch.utils.tensorboard.writer.SummaryWriter] instance for a given
    writer name.
    """
    def __init__(self, logdir: str) -> None:
        """Initialize the TensorBoard logger.

        Parameters:
          logdir: The root directory where the TensorBoard logs will be stored.
        """
        self._logdir: str = logdir
        self._writers: dict[str, torch.utils.tensorboard.writer.SummaryWriter] = {}

    def close(self) -> None:
        for writer in self._writers.values():
            writer.close()
        self._writers.clear()

    def get_writer(self, name: str) -> torch.utils.tensorboard.writer.SummaryWriter:
        """Possibly create and return a TensorBoard writer for the given name.

        Returns:
          writer: The opened TensorBoard writer.
        """
        if name not in self._writers:
            self._writers[name] = torch.utils.tensorboard.SummaryWriter(os.path.join(self._logdir, name))
        return self._writers[name]

    def _get_writer_from_label(self, label: str) -> tuple[torch.utils.tensorboard.writer.SummaryWriter, str]:
        """Possibly create and return a TensorBoard writer for the given label.

        Returns:
          writer: The opened TensorBoard writer.
          label: The label without the writer prefix.
        """
        writer, label = label.split(":", maxsplit=1) if ":" in label else ("train", label)
        return self.get_writer(writer), label

    def log_audio(self, label: str, audio: AnyArray, sample_rate: int, epoch: int) -> Self:
        audio = self.preprocess_audio(audio)
        audio = (audio.to(torch.float32) / 32_767).clamp(-1.0, 1.0).movedim(-1, 0)
        if audio.shape[0] == 2:
            audio = audio.mean(dim=0, keepdim=True)
        writer, label = self._get_writer_from_label(label)
        writer.add_audio(label, audio, epoch, sample_rate=sample_rate)
        writer.flush()
        return self

    def log_config(self, config: dict[str, Any], epoch: int) -> Self:
        writer = self.get_writer("train")
        writer.add_text("config", self.format_config_as_json(config), epoch)
        writer.flush()
        return self

    def log_figure(self, label: str, figure: Any, epoch: int, tight_layout: bool = True, close: bool = True) -> Self:
        return super().log_figure(label, figure, epoch, tight_layout, close)

    def log_graph(self, graph: torch.nn.Module, data: TensorOrTensors, epoch: int) -> Self:
        writer = self.get_writer("train")
        writer.add_graph(graph, data, use_strict_trace=False)
        writer.flush()
        return self

    def log_image(self, label: str, image: AnyArray, epoch: int, data_format: DataFormat = "HWC") -> Self:
        image = self.preprocess_image(image, data_format)
        writer, label = self._get_writer_from_label(label)
        writer.add_image(label, image, epoch, dataformats="HWC" if image.ndim == 3 else "HW")
        writer.flush()
        return self

    def log_metrics(self, logs: dict[str, float], epoch: int, description: str | None = None) -> Self:
        for label, value in logs.items():
            writer, label = self._get_writer_from_label(label)
            writer.add_scalar(label, value, epoch)

        for writer in self._writers.values():
            writer.flush()

        return self

    def log_text(self, label: str, text: str, epoch: int) -> Self:
        writer, label = self._get_writer_from_label(label)
        writer.add_text(label, text, epoch)
        writer.flush()
        return self



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import contextlib
import json
from typing import Any, Self

import numpy as np
import torch

from ..logger import Logger
from ..type_aliases import AnyArray, DataFormat


class BaseLogger(Logger):
    """An abstract logger providing base functionality for other loggers."""

    def log_figure(self, label: str, figure: Any, epoch: int, tight_layout: bool = True, close: bool = True) -> Self:
        import matplotlib.pyplot as plt
        import matplotlib.backends.backend_agg as plt_backend_agg

        tight_layout and figure.tight_layout()
        canvas = plt_backend_agg.FigureCanvasAgg(figure)
        canvas.draw()
        width, height = figure.canvas.get_width_height()
        image = torch.frombuffer(canvas.buffer_rgba(), dtype=torch.uint8).view(height, width, 4)
        close and plt.close(figure)

        return self.log_image(label, image, epoch)

    @staticmethod
    def format_config_as_json(config: dict[str, Any]) -> str:
        """Make a formatted JSON from configuration and epoch number."""
        return json.dumps(dict(sorted(config.items())), ensure_ascii=False, indent=2)

    @staticmethod
    def format_config_as_text(config: dict[str, Any], epoch: int) -> str:
        """Make a human-readable plain text from configuration and epoch number."""
        return " ".join(
            [f"Config epoch={epoch}"]
            + [f"{k}={v}" for k, v in sorted(config.items())]
        )

    @staticmethod
    def format_metrics(logs: dict[str, float]) -> str:
        """Make a human-readable string from the logged metrics."""
        return " ".join([f"{k}={v:#.{0 < abs(v) < 2e-4 and '2e' or '4f'}}" for k, v in logs.items()])

    @contextlib.contextmanager
    def graph_in_eval_mode(self, graph: torch.nn.Module):
        """Context manager to temporarily set the training mode of ``graph`` to eval."""
        if not isinstance(graph, torch.jit.ScriptFunction):
            old_training = graph.training
            graph.train(False)
            try:
                yield
            finally:
                graph.train(old_training)
        else:
            yield  # do nothing for a ScriptFunction

    @staticmethod
    def preprocess_audio(audio: AnyArray) -> torch.Tensor:
        """Produce a CPU-based [torch.Tensor][] with `dtype=torch.int16` and shape `(L, {1/2})`."""
        audio = torch.as_tensor(audio, device="cpu")
        audio = audio * 32_767 if audio.dtype.is_floating_point else audio
        audio = audio.clamp(-32_768, 32_767).to(torch.int16)
        assert audio.ndim == 1 or (audio.ndim == 2 and audio.shape[1] in (1, 2)), \
            "Audio must have shape (L,) or (L, 1/2)"
        if audio.ndim == 1:
            audio = audio.unsqueeze(-1)
        return audio

    @staticmethod
    def preprocess_image(image: AnyArray, data_format: DataFormat = "HWC") -> torch.Tensor:
        """Produce a CPU-based [torch.Tensor][] with `dtype=torch.uint8` and shape `(H, W, {1/3/4})`."""
        if type(image).__module__ == "PIL.Image":
            image, data_format = np.array(image, copy=True), "HWC"
        image = torch.as_tensor(image, device="cpu")
        image = image.movedim(0, -1) if data_format == "CHW" and image.ndim == 3 else image
        image = (image * 255 if image.dtype.is_floating_point else image).clamp(0, 255).to(torch.uint8)
        assert image.ndim == 2 or (image.ndim == 3 and image.shape[2] in (1, 2, 3, 4)), \
            "Image must have shape (H, W) or (H, W, 1/2/3/4)"
        if image.ndim == 2:
            image = image.unsqueeze(-1)
        if image.shape[2] == 2:
            # Convert to RGBA
            image = torch.stack([image[:, :, 0]] * 3 + [image[:, :, 1]], dim=-1)
        return image



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import html
from typing import Any, Self

import torch

from .base_logger import BaseLogger
from ..metric import Metric
from ..type_aliases import AnyArray, DataFormat, TensorOrTensors


class WandBLogger(BaseLogger):
    """A W&B logger interface.

    The text values are by default also logged as HTML for better visualization.
    """
    def __init__(self, project: str, *, text_also_as_html: bool = True, **kwargs: Any) -> None:
        """Create the WandBLogger with the given project name.

        Additional keyword arguments are passed to `wandb.init()`.

        Parameters:
          project: The name of the W&B project.
          text_also_as_html: Whether to log text messages also as HTML.
            That has the advantage of interactive visualization of the value
            at different epochs and preserving whitespace formatting.
          kwargs: Additional keyword arguments passed to `wandb.init()`, for example:

            - `dir` ([`str`][str]) – The directory where the W&B files will be stored in a `wandb` subdirectory.
            - `id` ([`str`][str]) – A unique identifier for this run.
            - `name` ([`str`][str]) – A short display name for this run to use in the UI.
            - `notes` ([`str`][str]) – A detailed description of the run.
            - `tags` ([`Sequence`][collections.abc.Sequence][[`str`][str]]) – A list of tags
              to label this run in the UI.
        """
        import wandb
        self.wandb = wandb
        self.run = self.wandb.init(project=project, **kwargs)
        self._text_also_as_html = text_also_as_html

    def close(self) -> None:
        if self.run:
            self.run.finish()
            self.run = None

    def _maybe_as_html(self, label: str, text: str) -> dict[str, Any]:
        """Return a dict with the HTML version of the text if enabled.

        The text is converted to HTML-safe format and returned as a wandb.Html object.
        """
        if not self._text_also_as_html:
            return {}
        return {f"{label}_html": self.wandb.Html("<pre>" + html.escape(text) + "</pre>")}

    def log_audio(self, label: str, audio: AnyArray, sample_rate: int, epoch: int) -> Self:
        audio = self.preprocess_audio(audio).numpy()
        self.run.log({label: self.wandb.Audio(audio, sample_rate=sample_rate)}, step=epoch)
        return self

    def log_config(self, config: dict[str, Any], epoch: int) -> Self:
        self.run.config.update(dict(sorted(config.items())))
        config_json = self.format_config_as_json(config)
        self.run.log({"config": config_json} | self._maybe_as_html("config", config_json), step=epoch)
        return self

    def log_figure(self, label: str, figure: Any, epoch: int, tight_layout: bool = True, close: bool = True) -> Self:
        return super().log_figure(label, figure, epoch, tight_layout, close)

    def log_graph(self, graph: torch.nn.Module, data: TensorOrTensors, epoch: int) -> Self:
        # The logging in WandB has a lot of limitations. One is that all
        # children of the graph must be executed before the graph is logged;
        # however, metrics are usually not executed during the forward pass.
        # We therefore try to exclude metrics from the logged graph.
        watched_children = torch.nn.Module()
        for name, child in graph.named_children():
            if not any(isinstance(module, Metric) for module in child.modules()):
                watched_children.add_module(name, child)
        self.run.watch(watched_children, log=None, log_graph=True)
        with self.graph_in_eval_mode(graph), torch.no_grad():
            graph(*data) if isinstance(data, (list, tuple)) else graph(data)  # Run the graph to log it.
        self.run.unwatch(watched_children)
        return self

    def log_image(self, label: str, image: AnyArray, epoch: int, data_format: DataFormat = "HWC") -> Self:
        image = self.preprocess_image(image, data_format).numpy()
        self.run.log({label: self.wandb.Image(image)}, step=epoch)
        return self

    def log_metrics(self, logs: dict[str, float], epoch: int, description: str | None = None) -> Self:
        self.run.log(logs, step=epoch)
        return self

    def log_text(self, label: str, text: str, epoch: int) -> Self:
        self.run.log({label: text} | self._maybe_as_html(label, text), step=epoch)
        return self



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from collections.abc import Iterable
from typing import Any, Self

import torch

from .mean import Mean


class ExactMatch(Mean):
    """Exact match metric implementation.

    The elements to compare can be either tensors or generic iterables. When tensors are used,
    the `element_dims` parameter can be specified to indicate which dimensions of the tensors
    form an element for comparison; when iterables are used, the input one-dimensional sequences are
    compared directly.
    """

    def __init__(self, element_dims: int | tuple[int] = (), device: torch.device | None = None) -> None:
        """Create the ExactMatch metric object.

        Parameters:
          element_dims: If the values to compare are tensors, this parameter
            can be used to specify which dimensions of the tensors form
            an element for comparison.
        """
        super().__init__(device)
        if isinstance(element_dims, int):
            self._element_dims = (element_dims,)
        elif isinstance(element_dims, (tuple, list)):
            self._element_dims = tuple(element_dims)
        else:
            raise TypeError("The element_dims argument must be an int or a tuple of ints.")

    @torch.no_grad
    def update(
        self,
        y: torch.Tensor | Iterable[Any],
        y_true: torch.Tensor | Iterable[Any],
        sample_weights: torch.Tensor | None = None,
    ) -> Self:
        """Update the exact match by comparing the given values.

        The inputs can be either both tensors or both iterables. When they are both tensors,
        the `element_dims` parameter can be used to specify which dimensions of the tensors
        form an element for comparison; when they are both iterables, the elements are
        compared directly.

        Optional sample weight might be provided; if not, all values are weighted with 1.

        Parameters:
          y: A tensor or an iterable of predicted values of the same shape as `y_true`.
          y_true: A tensor or an iterable of ground-truth targets of the same shape as `y`.
          sample_weights: Optional sample weights. Their shape must be broadcastable to a
            prefix of the shape of `y` (with `element_dims` dimensions removed, if specified).

        Returns:
          self
        """
        if isinstance(y, torch.Tensor) and isinstance(y_true, torch.Tensor):
            assert y.shape == y_true.shape, "The y and y_true tensors must have the same shape."

            equals = (y == y_true)
            if self._element_dims:
                equals = torch.all(equals, dim=self._element_dims)
        else:
            assert not self._element_dims, "Nonempty element_dims can only be used with tensor inputs."
            equals = torch.tensor([pred == true for pred, true in zip(y, y_true, strict=True)], dtype=torch.float32)

        return super().update(equals, sample_weights=sample_weights)



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from collections.abc import Callable
import math
from typing import Any

import torch


class KerasParameterInitialization:
    def reset_parameters_linear(self) -> None:
        torch.nn.init.xavier_uniform_(self.weight)
        if self.bias is not None:
            torch.nn.init.zeros_(self.bias)

    def reset_parameters_bilinear(self) -> None:
        # Keras does not have a Bilinear layer. But we analogously use
        # the Xavier uniform initialization, where
        # - the fan_out for each out_feature is in_feature1 * in_feature2
        # - the fan_in for each in_feature1 is out_feature * in_feature2
        # - the fan_in for each in_feature2 is out_feature * in_feature1
        # - the overall fan_in is computed as a weighted average of the above two as
        #   (2 * out_feature * in_feature1 * in_feature2) / (in_feature1 + in_feature2)
        out, in1, in2 = self.weight.shape
        fan_in = (2 * out * in1 * in2) / (in1 + in2)
        fan_out = in1 * in2
        bound = math.sqrt(6 / (fan_in + fan_out))
        torch.nn.init.uniform_(self.weight, -bound, bound)
        if self.bias is not None:
            torch.nn.init.zeros_(self.bias)

    def reset_parameters_rnn(self) -> None:
        for name, parameter in self.named_parameters():
            if "weight_ih" in name:
                torch.nn.init.xavier_uniform_(parameter)
            elif "weight_hh" in name:
                torch.nn.init.orthogonal_(parameter)
            elif "bias" in name:
                torch.nn.init.zeros_(parameter)
                if isinstance(self, (torch.nn.LSTM, torch.nn.LSTMCell)):  # Set LSTM forget gate bias to 1
                    parameter.data[self.hidden_size:self.hidden_size * 2] = 1

    def reset_parameters_embedding(self) -> None:
        torch.nn.init.uniform_(self.weight, -0.05, 0.05)
        self._fill_padding_idx_with_zero()

    overrides: dict[torch.nn.Module, Callable] = {
        torch.nn.Linear: reset_parameters_linear,
        torch.nn.Conv1d: reset_parameters_linear,
        torch.nn.Conv2d: reset_parameters_linear,
        torch.nn.Conv3d: reset_parameters_linear,
        torch.nn.ConvTranspose1d: reset_parameters_linear,
        torch.nn.ConvTranspose2d: reset_parameters_linear,
        torch.nn.ConvTranspose3d: reset_parameters_linear,
        torch.nn.Bilinear: reset_parameters_bilinear,
        torch.nn.RNN: reset_parameters_rnn,
        torch.nn.RNNCell: reset_parameters_rnn,
        torch.nn.LSTM: reset_parameters_rnn,
        torch.nn.LSTMCell: reset_parameters_rnn,
        torch.nn.GRU: reset_parameters_rnn,
        torch.nn.GRUCell: reset_parameters_rnn,
        torch.nn.Embedding: reset_parameters_embedding,
        torch.nn.EmbeddingBag: reset_parameters_embedding,
    }


class KerasNormalizationLayers:
    @staticmethod
    def override_default_argument_value(func: Callable, name: str, default: Any) -> None:
        default_names = func.__code__.co_varnames[:func.__code__.co_argcount][-len(func.__defaults__):]
        assert name in default_names, f"Argument {name} not found in {func.__name__} arguments"
        func.__defaults__ = tuple(
            default if arg_name == name else arg_value
            for arg_name, arg_value in zip(default_names, func.__defaults__)
        )

    batch_norms = [
        torch.nn.BatchNorm1d,
        torch.nn.BatchNorm2d,
        torch.nn.BatchNorm3d,
        torch.nn.LazyBatchNorm1d,
        torch.nn.LazyBatchNorm2d,
        torch.nn.LazyBatchNorm3d,
        torch.nn.SyncBatchNorm,
    ]

    all_norms = batch_norms + [
        torch.nn.LayerNorm,
        torch.nn.GroupNorm,
    ]


def global_keras_initializers(
    parameter_initialization: bool = True,
    batchnorm_momentum_override: float | None = 0.01,
    norm_layer_epsilon_override: float | None = 0.001,
) -> None:
    """Change default PyTorch initializers to Keras defaults.

    The following initializers are used:

    - `Linear`, `Conv1d`, `Conv2d`, `Conv3d`, `ConvTranspose1d`, `ConvTranspose2d`, `ConvTranspose3d`, `Bilinear`:
      Xavier uniform for weights, zeros for biases.
    - `Embedding`, `EmbeddingBag`: Uniform [-0.05, 0.05] for weights.
    - `RNN`, `RNNCell`, `LSTM`, `LSTMCell`, `GRU`, `GRUCell`: Xavier uniform for input weights,
      orthogonal for recurrent weights, zeros for biases (with LSTM forget gate bias set to 1).

    Furthermore, for batch normalization layers, the default momentum value is changed
    from 0.1 to the Keras default of 0.01 (or any other value specified).

    Finally, for batch normalization, layer normalization, and group normalization layers,
    the default epsilon value is changed from 1e-5 to the Keras default of 1e-3
    (or any other value specified).

    Parameters:
     parameter_initialization: If True, override the default PyTorch initializers with Keras defaults.
     batchnorm_momentum_override: If not None, override the default value of batch normalization
       momentum from 0.1 to this value.
     norm_layer_epsilon_override: If not None, override the default value of epsilon
       for batch normalization, layer normalization, and group normalization layers from
       1e-5 to this value.
    """
    if parameter_initialization:
        for class_, reset_parameters_method in KerasParameterInitialization.overrides.items():
            class_.reset_parameters = reset_parameters_method

    if batchnorm_momentum_override is not None:
        for batch_norm_super in KerasNormalizationLayers.batch_norms:
            for batch_norm in [batch_norm_super] + batch_norm_super.__subclasses__():
                KerasNormalizationLayers.override_default_argument_value(
                    batch_norm.__init__, "momentum", batchnorm_momentum_override
                )

    if norm_layer_epsilon_override is not None:
        for norm_layer_super in KerasNormalizationLayers.all_norms:
            for norm_layer in [norm_layer_super] + norm_layer_super.__subclasses__():
                KerasNormalizationLayers.override_default_argument_value(
                    norm_layer.__init__, "eps", norm_layer_epsilon_override
                )



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Types and type aliases used by NPFL138."""
from typing import Any, Literal, Protocol, TypeAlias

import numpy as np
import torch


AnyArray: TypeAlias = torch.Tensor | np.ndarray | list | tuple
"""A type alias for any array-like structure.

PyTorch tensors, NumPy arrays, lists, and tuples are supported.
"""


DataFormat: TypeAlias = Literal["HWC", "CHW"]
"""A type alias for image data format description."""


class HasCompute(Protocol):
    """A protocol for objects that have a `compute` method of the following type:

    ```python
    def compute(self) -> float | torch.Tensor | np.ndarray:
        ...
    ```
    """
    def compute(self) -> float | torch.Tensor | np.ndarray:
        """Compute the value of the object."""
        ...


Logs: TypeAlias = dict[str, float | torch.Tensor | np.ndarray | HasCompute]
"""A dictionary of logs, with keys being the log names and values being the log values.

When the logs are returned by a [npfl138.TrainableModule][] or passed to a [npfl138.Callback][],
they are always evaluated to just float values.
"""

Reduction: TypeAlias = Literal["mean", "sum", "none"]
"""A type alias for reduction methods used in losses and metrics."""


Tensor: TypeAlias = torch.Tensor | torch.nn.utils.rnn.PackedSequence
"""A type alias for a single tensor or a packed sequence of tensors."""


TensorOrTensors: TypeAlias = Tensor | tuple[Tensor, ...] | list[Tensor] | dict[str, Tensor] | Any
"""A type alias for a single tensor or a tensor structure.

While a tensor or a sequence of them is the most common, any type is allowed
here to accomodate nested or completely custom data structures.
"""



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import TYPE_CHECKING

from ..callback import Callback
from ..type_aliases import Logs
if TYPE_CHECKING:
    from ..trainable_module import TrainableModule


class SaveWeights(Callback):
    """A callback that saves model weights to a file after each epoch."""

    def __init__(self, path: str, optimizer_path: str | None = None) -> None:
        """Create the SaveWeights callback.

        Parameters:
          path: A path where weights will be saved using the [npfl138.TrainableModule.save_weights][]
            method after each epoch. Note that you can use templates like `{logdir}` and `{epoch[:formatting]}`.
          optimizer_path: An optional path passed to [npfl138.TrainableModule.save_weights][] to
            save also the optimizer state; it is relative to `path`.
        """
        self._path = path
        self._optimizer_path = optimizer_path

    def __call__(self, module: "TrainableModule", epoch: int, logs: Logs) -> None:
        module.save_weights(self._path, optimizer_path=self._optimizer_path)



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import os
import random

import numpy as np
import torch


def startup(
    seed: int | None = None,
    threads: int | None = None,
    recodex: bool = False,
    *,
    forkserver_instead_of_fork: bool = True,
    allow_tf32: bool = True,
    expandable_segments: bool | None = True,
) -> None:
    """Initialize the environment.

    - Set the random seed if given.
    - Set the number of threads if given.
    - Use `forkserver` instead of `fork` multiprocessing start method unless disallowed.
    - Allow using TF32 for matrix multiplication unless disallowed.
    - Enable expandable segments in the CUDA memory allocator unless disallowed.

    Parameters:
      seed: If not `None`, set the Python, Numpy, and PyTorch random seeds to this value.
      threads: If not `None` of 0, set the number of threads to this value.
        Otherwise, use as many threads as cores.
      recodex: If `True`, run in ReCodEx mode for better replicability of tests.
      forkserver_instead_of_fork: If `True`, use `forkserver` instead of `fork` as the
        default start multiprocessing method. This will be the default one in Python 3.14.
      allow_tf32: If `False`, disable TF32 for matrix multiplication even when available.
      expandable_segments: If `True`, enable expandable segments in the CUDA memory allocator;
        if `False`, disable them; if `None`, do not change the current setting.

    **Environment variables:** The following environment variables can be used
    to override the method parameters:

    - `NPFL_START_METHOD`: If set to `fork` or `forkserver`, uses the specified method as
      the multiprocessing start method.
    - `NPFL_ALLOW_TF32`: If set to `0` or `1`, overrides the `allow_tf32` parameter.
    - `NPFL_EXPANDABLE_SEGMENTS`: If set to `0` or `1`, overrides the `expandable_segments` parameter.
    """

    # Set random seed if not None.
    if seed is not None:
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)

    # Set number of threads if > 0; otherwise, use as many threads as cores.
    if threads is not None and threads > 0:
        if torch.get_num_threads() != threads:
            torch.set_num_threads(threads)
        if torch.get_num_interop_threads() != threads:
            torch.set_num_interop_threads(threads)

    # If instructed, use `forkserver` instead of `fork` (which will be the default in Python 3.14).
    if "fork" in torch.multiprocessing.get_all_start_methods():
        if os.environ.get("NPFL_START_METHOD") == "fork":
            if torch.multiprocessing.get_start_method(allow_none=True) != "fork":
                torch.multiprocessing.set_start_method("fork")
        elif forkserver_instead_of_fork or os.environ.get("NPFL_START_METHOD") == "forkserver":
            if torch.multiprocessing.get_start_method(allow_none=True) != "forkserver":
                torch.multiprocessing.set_start_method("forkserver")

    # Allow TF32 for matrix multiplication if available, unless instructed otherwise.
    if os.environ.get("NPFL_ALLOW_TF32") in ["0", "1"]:
        allow_tf32 = os.environ.get("NPFL_ALLOW_TF32") == "1"
    torch.backends.cuda.matmul.allow_tf32 = allow_tf32

    # On NVIDIA GPUs, allow or disallow expandable segments in the CUDA memory allocator if requested.
    if os.environ.get("NPFL_EXPANDABLE_SEGMENTS") in ["0", "1"]:
        expandable_segments = os.environ.get("NPFL_EXPANDABLE_SEGMENTS") == "1"
    if expandable_segments is not None:
        expandable_segments = bool(expandable_segments)
        env_variables = [os.environ.get(k, "") for k in ("PYTORCH_ALLOC_CONF", "PYTORCH_CUDA_ALLOC_CONF")]
        if not any(f"expandable_segments:{str(not expandable_segments)}" in env_var for env_var in env_variables):
            if torch.cuda.is_available() and torch.version.cuda:
                # Since PyTorch 2.10, we need to use the accelerator API instead of CUDA to set expandable segments.
                set_allocator_settings = getattr(torch._C, "_accelerator_setAllocatorSettings", None)
                set_allocator_settings = set_allocator_settings or torch.cuda.memory._set_allocator_settings
                set_allocator_settings(f"expandable_segments:{str(expandable_segments)}")

    # If ReCodEx mode is requested, apply various overrides for better replicability.
    if recodex:
        # Do not use accelerators.
        torch.cuda.is_available = lambda: False
        torch.backends.mps.is_available = lambda: False
        torch.xpu.is_available = lambda: False

        # Make initializers deterministic.
        def bind_generator(init):
            return lambda *args, **kwargs: init(*args, **kwargs | {"generator": torch.Generator().manual_seed(seed)})
        for name in ["uniform_", "normal_", "trunc_normal_", "xavier_uniform_", "xavier_normal_",
                     "kaiming_uniform_", "kaiming_normal_", "orthogonal_", "sparse_"]:
            setattr(torch.nn.init, name, bind_generator(getattr(torch.nn.init, name)))

        # Override the generator of every DataLoader to a fresh default generator.
        original_dataloader_init = torch.utils.data.DataLoader.__init__
        torch.utils.data.DataLoader.__init__ = lambda self, dataset, *args, **kwargs: original_dataloader_init(
            self, dataset, *args, **kwargs | {"generator": torch.Generator().manual_seed(seed)})



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""The `GymCartpoleDataset` is a toy dataset with 100 samples from the CartPole environment.

The dataset consists of just a single subset:

- `train`: 100 examples, each a `(observation, label)` pair, where
    - `inputs` is a vector with [npfl138.datasets.gym_cartpole_dataset.GymCartpoleDataset.FEATURES][]
      floating point values,
    - `label` is a gold 0/1 class index.
"""
import collections
import os
import sys
import urllib.request

import numpy as np
import torch


class GymCartpoleDataset:
    FEATURES: int = 4
    """Number of features of every example in the dataset."""

    Element = collections.namedtuple("Element", ["observation", "label"])
    """The type of a single dataset element."""

    URL: str = "https://ufal.mff.cuni.cz/~straka/courses/npfl138/2526/datasets/"

    class Dataset(torch.utils.data.Dataset):
        def __init__(self, observations: np.ndarray, labels: np.ndarray) -> None:
            self._observations = torch.as_tensor(observations)
            self._labels = torch.as_tensor(labels)

        def __len__(self) -> int:
            """Return the number of elements in the dataset."""
            return len(self._observations)

        def __getitem__(self, index: int) -> "GymCartpoleDataset.Element":
            """Return the `index`-th element of the dataset."""
            return GymCartpoleDataset.Element(self._observations[index], self._labels[index])

        @property
        def observations(self) -> torch.Tensor:
            """All dataset observations as a single tensor."""
            return self._observations

        @property
        def labels(self) -> torch.Tensor:
            """All dataset labels as a single tensor."""
            return self._labels

    def __init__(self, dataset: str = "gym_cartpole_data.txt") -> None:
        """Load the GymCartpoleDataset, downloading it if necessary.

        Parameters:
          dataset: The file name of the dataset to load.
        """
        if not os.path.exists(dataset):
            print("Downloading {}...".format(dataset), file=sys.stderr)
            urllib.request.urlretrieve("{}/{}".format(self.URL, dataset), filename="{}.tmp".format(dataset))
            os.rename("{}.tmp".format(dataset), dataset)

        data = np.loadtxt(dataset)
        self.train = self.Dataset(observations=data[:, :-1].astype(np.float32), labels=data[:, -1].astype(np.int64))

    train: Dataset
    """The training dataset."""



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# TrainableModule
from .trainable_module import TrainableModule

# TransformedDataset
from .transformed_dataset import TransformedDataset

# Type aliases
from .type_aliases import AnyArray, DataFormat, HasCompute, Logs, Reduction, Tensor, TensorOrTensors

# Vocabulary
from .vocabulary import Vocabulary

# Utils
from .first_time import first_time
from .format_logdir_impl import format_logdir
from .initializers_override import global_keras_initializers
from .progress_logger import ProgressLogger
from .startup_impl import startup
from .trainable_module import tensors_to_device
from .version import require_version, __version__

# Callbacks
from .callback import Callback, STOP_TRAINING
from . import callbacks

# Loggers
from .logger import Logger
from . import loggers

# Losses
from .loss import Loss
from . import losses

# Metrics
from .metric import Metric
from . import metrics



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

__version__ = "2526.2.0"


def require_version(required_version: str) -> None:
    """Make sure the installed version is at least `required_version`.

    If not, show a nice error message with the instructions how to upgrade.

    Parameters:
      required_version: The required version of the npfl138 package, as
        a string in the format "semester.week" or "semester.week.patch".
    """

    required = required_version.split(".")
    assert len(required) <= 3, "Expected at most 3 version components"

    required = list(map(int, required))
    current = list(map(int, __version__.split(".")))

    assert current[:len(required)] >= required, (
        f"The npfl138>={required_version} is required, but found only {__version__}.\n"
        f"Please update the npfl138 package by running either:\n"
        f"- `VENV_DIR/bin/pip install --upgrade npfl138` when using a venv, or\n"
        f"- `python3 -m pip install --user --upgrade npfl138` otherwise.")



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import gc
import os
import struct
from typing import Any, Self, TextIO
import wave
import zlib

import torch

from .base_logger import BaseLogger
from ..type_aliases import AnyArray, DataFormat, TensorOrTensors
from ..utils import sanitize_path


class FileSystemLogger(BaseLogger):
    """A file system logger interface."""

    def __init__(self, logdir: str) -> None:
        """Initialize the file system logger.

        Parameters:
          logdir: The root directory where the log files will be stored.
        """
        self._logdir: str = logdir
        self._log_file: TextIO | None = None

    def close(self) -> None:
        if self._log_file:
            self._log_file.close()
            self._log_file = None

    def get_file(self) -> TextIO:
        """Possibly open and return log file object.

        Returns:
          file: The opened log file.
        """
        if not self._log_file:
            os.makedirs(self._logdir, exist_ok=True)
            self._log_file = open(os.path.join(self._logdir, "logs.txt"), "a", encoding="utf-8")

        return self._log_file

    def _split_label(self, label: str) -> tuple[str, str]:
        """Split the given label into directory and base label, and sanitize them.

        Returns:
          directory: The directory part of the label.
          label: The base label.
        """
        directory, label = label.split(":", maxsplit=1) if ":" in label else ("train", label)
        directory, label = map(sanitize_path, (directory, label))
        return os.path.join(self._logdir, directory), label

    def _maybe_epoch(self, epoch: int) -> str:
        """Return epoch suffix if epoch is non-zero.

        Returns:
          suffix: The epoch suffix.
        """
        return f".{epoch}" if epoch else ""

    def log_audio(self, label: str, audio: AnyArray, sample_rate: int, epoch: int) -> Self:
        audio = self.preprocess_audio(audio)

        directory, label = self._split_label(label)
        os.makedirs(directory, exist_ok=True)
        with wave.open(os.path.join(directory, f"{label}{self._maybe_epoch(epoch)}.wav"), "wb") as wav_file:
            wav_file.setsampwidth(2)  # 16 bits
            wav_file.setnchannels(audio.shape[-1])
            wav_file.setframerate(sample_rate)
            wav_file.writeframes(audio.numpy().tobytes())

        return self

    def log_config(self, config: dict[str, Any], epoch: int) -> Self:
        print(self.format_config_as_text(config, epoch), file=self.get_file(), flush=True)
        return self

    def log_figure(self, label: str, figure: Any, epoch: int, tight_layout: bool = True, close: bool = True) -> Self:
        return super().log_figure(label, figure, epoch, tight_layout, close)

    def log_graph(self, graph: torch.nn.Module, data: TensorOrTensors, epoch: int) -> Self:
        os.makedirs(self._logdir, exist_ok=True)
        with open(os.path.join(self._logdir, f"graph{self._maybe_epoch(epoch)}.txt"), "w", encoding="utf-8") as file:
            if isinstance(graph, torch.nn.Sequential):
                print("# Sequential Module", graph, file=file, sep="\n", end="\n\n")
            with self.graph_in_eval_mode(graph), torch.no_grad():
                traced = torch.jit.trace(graph, data, check_trace=False, strict=False)
            print("# Traced Code", traced.code, file=file, sep="\n", end="\n\n")
            print("# Traced Graph", traced.graph, file=file, sep="\n")
            print("# Traced Inlined Graph", traced.inlined_graph, file=file, sep="\n")
        del traced
        gc.collect()  # Make sure the traced module is collected.
        return self

    def log_image(self, label: str, image: AnyArray, epoch: int, data_format: DataFormat = "HWC") -> Self:
        image = self.preprocess_image(image, data_format).numpy()

        directory, label = self._split_label(label)
        os.makedirs(directory, exist_ok=True)
        with open(os.path.join(directory, f"{label}{self._maybe_epoch(epoch)}.png"), "wb") as file:
            def add_block(chunk_type: bytes, data: bytes) -> None:
                file.write(struct.pack("!I", len(data)))
                data = chunk_type + data
                file.write(data)
                file.write(struct.pack("!I", zlib.crc32(data)))

            file.write(b"\x89PNG\r\n\x1a\n")
            add_block(b"IHDR", struct.pack(
                "!2I5B", image.shape[1], image.shape[0], 8, [0, 4, 2, 6][image.shape[2] - 1], 0, 0, 0))
            pixels = [b"\2" + (image[y] - (image[y - 1] if y else 0)).tobytes() for y in range(len(image))]
            add_block(b"IDAT", zlib.compress(b"".join(pixels), level=9))
            add_block(b"IEND", b"")

        return self

    def log_metrics(self, logs: dict[str, float], epoch: int, description: str | None = None) -> Self:
        print(description or f"Epoch {epoch}", self.format_metrics(logs), file=self.get_file(), flush=True)
        return self

    def log_text(self, label: str, text: str, epoch: int) -> Self:
        directory, label = self._split_label(label)
        os.makedirs(directory, exist_ok=True)
        with open(os.path.join(directory, f"{label}{self._maybe_epoch(epoch)}.txt"), "w", encoding="utf-8") as file:
            file.write(text)
        return self



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Literal, TYPE_CHECKING

from ..callback import Callback
from ..type_aliases import Logs
if TYPE_CHECKING:
    from ..trainable_module import TrainableModule


class SaveBestWeights(Callback):
    """A callback that saves best model weights to a file."""

    def __init__(
        self,
        path: str,
        metric: str,
        mode: Literal["max", "min"] = "max",
        optimizer_path: str | None = None,
    ) -> None:
        """Create the SaveBestWeights callback.

        Parameters:
          path: A path where weights will be saved using the [npfl138.TrainableModule.save_weights][]
            method after each epoch. Note that you can use templates like `{logdir}` and `{epoch[:formatting]}`.
          metric: The metric name from `logs` dictionary to monitor.
          mode: One of `"max"` or `"min"`, indicating whether the monitored metric should be maximized
            or minimized.
          optimizer_path: An optional path passed to [npfl138.TrainableModule.save_weights][] to
            save also the optimizer state; it is relative to `path`.
        """
        assert mode in ("max", "min"), "mode must be one of 'max' or 'min'"

        self._path = path
        self._metric = metric
        self._mode = mode
        self._optimizer_path = optimizer_path

        self.best_metric_value = None

    best_value: float | None = None
    """The best metric value seen so far."""

    def __call__(self, module: "TrainableModule", epoch: int, logs: Logs) -> None:
        if (self.best_value is None
                or (self._mode == "max" and logs[self._metric] > self.best_value)
                or (self._mode == "min" and logs[self._metric] < self.best_value)):
            self.best_value = logs[self._metric]

            module.save_weights(self._path, optimizer_path=self._optimizer_path)



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import torch

from ..loss import Loss
from ..type_aliases import Reduction
from ..utils import broadcast_to_prefix, maybe_remove_one_singleton_dimension


class BinaryCrossEntropy(Loss):
    """Binary cross-entropy loss implementation."""

    def __init__(self, *, label_smoothing: float = 0.0, probs: bool = False, reduction: Reduction = "mean") -> None:
        """Create the BinaryCrossEntropy loss object with the specified reduction method.

        Parameters:
          label_smoothing: A float in [0.0, 1.0] specifying the label smoothing factor.
            If greater than 0.0, the used ground-truth targets are computed as a mixture
            of the original targets and uniform distribution with weight `1 - label_smoothing`.
          probs: If `False`, the predictions are assumed to be logits; if `True`, the
            predictions are assumed to be probabilities. Note that gold targets are
            always expected to be probabilities.
          reduction: The reduction method to apply to the computed loss.
        """

        self._label_smoothing = label_smoothing
        self._reduction = reduction
        if probs:
            self._loss_fn = torch.nn.functional.binary_cross_entropy
        else:
            self._loss_fn = torch.nn.functional.binary_cross_entropy_with_logits

    def __call__(
        self, y: torch.Tensor, y_true: torch.Tensor, sample_weights: torch.Tensor | None = None,
    ) -> torch.Tensor:
        """Compute the binary cross-entropy loss, optionally with sample weights.

        Parameters:
          y: The predicted outputs. Their shape either has to be exactly the same as `y_true` (no broadcasting),
            or can contain an additional single dimension of size 1.
          y_true: The ground-truth targets.
          sample_weights: Optional sample weights. If provided, their shape must be broadcastable
            to a prefix of a shape of `y_true`, and the loss for each sample is weighted accordingly.

        Returns:
          A tensor representing the computed loss. A scalar tensor if reduction is `"mean"` or `"sum"`;
            otherwise (if reduction is `"none"`), a tensor of the same shape as `y_true`.
        """
        y = maybe_remove_one_singleton_dimension(y, y_true)
        assert y.shape == y_true.shape, f"Shapes of y {y.shape} and y_true {y_true.shape} have to match " \
            "up to one singleton dim in y."

        if self._label_smoothing:
            y_true = y_true * (1.0 - self._label_smoothing) + 0.5 * self._label_smoothing

        if sample_weights is not None:
            sample_weights = broadcast_to_prefix(sample_weights, y_true.shape)

        return self._loss_fn(y, y_true, reduction=self._reduction, weight=sample_weights)



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Any, Protocol

import torch


class Metric(Protocol):
    """An abstract metric interface.

    Note:
      Metrics are expected to be instances of [torch.nn.Module][] and are stored
        in a [torch.nn.ModuleDict][].
    """

    def update(self, y: torch.Tensor, y_true: torch.Tensor) -> Any:
        """Update the internal state of the metric with new predictions and gold targets.

        Parameters:
          y: The predicted outputs.
          y_true: The ground-truth targets.

        Returns:
          anything; `npfl138` metrics return `Self`, but any return value is allowed in the generic
            interface (`torchmetrics` metrics return `None`, for example).
        """
        ...

    def compute(self) -> torch.Tensor:
        """Compute the accumulated metric value.

        Returns:
          A (usually scalar) tensor representing the accumulated metric value.
        """
        ...

    def reset(self) -> Any:
        """Reset the internal state of the metric.

        Returns:
          anything; `npfl138` metrics return `Self`, but any return value is allowed in the generic
            interface (`torchmetrics` metrics return `None`, for example).
        """
        ...



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import os
import re
from typing import Any

import torch

from .type_aliases import Logs


def broadcast_to_prefix(tensor: torch.Tensor, shape: torch.Size) -> torch.Tensor:
    """Broadcast the given tensor to the given shape from the left (prefix).

    The broadcasting is performed by unsqueezing the tensor's dimensions until
    it has the same number of dimensions as the given shape, and then expanding it.

    Parameters:
      tensor: The tensor to broadcast.
      shape: The shape to broadcast the tensor to (from the left).

    Returns:
      The broadcasted tensor.
    """
    while tensor.dim() < len(shape):
        tensor = tensor.unsqueeze(dim=-1)
    if tensor.shape != shape:
        tensor = tensor.expand(shape)
    return tensor


def compute_logs(logs: Logs) -> dict[str, float]:
    """Evaluate the given logs dictionary to floats, calling `compute` where needed.

    Parameters:
      logs: The logs dictionary which can contain float values, PyTorch and Numpy tensors,
        or [HasCompute][]-compliant objects providing a `compute` method.

    Returns:
      logs: The same dictionary as on input, but with all values evaluated to floats.
    """
    for k, v in logs.items():
        if not isinstance(v, float):
            v = v.compute() if hasattr(v, "compute") else v
            logs[k] = float(v.item() if hasattr(v, "item") else v)
    return logs


def maybe_remove_one_singleton_dimension(y: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
    """Possibly remove one singleton dimension from the given tensor `y` to match the shape of `y_true`.

    If `y` has one more dimension than `y_true` and that extra dimension is a dimension of size 1,
    it will remove that dimension from `y`. Otherwise, `y` is returned unchanged.

    Parameters:
      y: The predicted outputs.
      y_true: The ground-truth targets.

    Returns:
      The tensor `y` with a surplus singleton dimension possibly removed.
    """
    y_shape, y_true_shape = y.shape, y_true.shape

    if len(y_shape) == len(y_true_shape) + 1:
        singleton_dim = 0
        while singleton_dim < len(y_true_shape) and y_shape[singleton_dim] == y_true_shape[singleton_dim]:
            singleton_dim += 1
        if y_shape[singleton_dim] == 1:
            y = y.squeeze(dim=singleton_dim)

    return y


def fill_and_standardize_path(path: str, **kwargs: Any) -> str:
    """Fill placeholders in the path and standardize path separators.

    The template placeholders `{key}` in the path are replaced with the corresponding values
    from `kwargs` using `str.format`, and the both slashes and backslashes are replaced
    with the current OS path separator.

    Parameters:
      path: The path template with placeholders.
      **kwargs: The keyword arguments to fill the placeholders in the path.

    Returns:
      The standardized path with filled placeholders and OS-specific separators.
    """
    filled_path = path.format(**kwargs)
    standardized_path = filled_path.replace("\\", os.path.sep).replace("/", os.path.sep)
    return standardized_path


def sanitize_path(path: str) -> str:
    """Sanitize the given path by replacing path-unfriendly characters with underscores.

    Parameters:
      path: The input path to sanitize.

    Returns:
      The sanitized path.
    """
    return re.sub(r'[<>:"/\\|?*\x00-\x1f]', "_", path)


tuple_list: tuple[type, type] = (tuple, list)



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Self

import torch

from .mean import Mean
from .. import losses


class MeanSquaredError(Mean):
    """Mean squared error metric implementation."""

    def __init__(self, device: torch.device | None = None) -> None:
        """Create the MeanSquaredError metric object."""
        super().__init__(device)
        self._mse_loss = losses.MeanSquaredError(reduction="none")

    @torch.no_grad
    def update(
        self, y: torch.Tensor, y_true: torch.Tensor, sample_weights: torch.Tensor | None = None,
    ) -> Self:
        """Update the accumulated mean squared error by introducing new values.

        Optional sample weight might be provided; if not, all values are weighted with 1.

        Parameters:
          y: The predicted outputs. Their shape either has to be exactly the same as `y_true` (no broadcasting),
            or can contain an additional single dimension of size 1.
          y_true: The ground-truth targets.
          sample_weights: Optional sample weights. If provided, their shape must be broadcastable
            to a prefix of a shape of `y_true`, and the loss for each sample is weighted accordingly.

        Returns:
          self
        """
        return super().update(self._mse_loss(y, y_true), sample_weights=sample_weights)



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from abc import ABC, abstractmethod
from typing import Any, Self

import torch

from .type_aliases import AnyArray, DataFormat, TensorOrTensors


class Logger(ABC):
    """An abstract logger interface for logging metrics and other information."""

    @abstractmethod
    def close(self) -> None:
        """Close the logger and release its resources."""
        ...

    @abstractmethod
    def log_audio(self, label: str, audio: AnyArray, sample_rate: int, epoch: int) -> Self:
        """Log the given audio with the given label at the given epoch.

        Parameters:
          label: The label of the logged audio.
          audio: The audio to log, represented as an array with any of the
            following shapes:

            - `(L,)` of `(L, 1)` for mono audio,
            - `(L, 2)` for stereo audio.

            If the sample values are floating-point numbers, they are expected
            to be in the `[-1, 1]` range; otherwise, they are assumed to be in the
            `[-32_768, 32_767]` range.
          sample_rate: The sample rate of the audio.
          epoch: The epoch number at which the audio is logged.
        """
        ...

    @abstractmethod
    def log_config(self, config: dict[str, Any], epoch: int) -> Self:
        """Log the given configuration dictionary at the given epoch.

        Parameters:
          config: A JSON-serializable dictionary representing the configuration to log.
          epoch: The epoch number at which the configuration is logged.
        """
        ...

    @abstractmethod
    def log_figure(self, label: str, figure: Any, epoch: int, tight_layout: bool = True, close: bool = True) -> Self:
        """Log the given matplotlib Figure with the given label at the given epoch.

        Parameters:
          label: The label of the logged image.
          figure: A matplotlib Figure.
          epoch: The epoch number at which the image is logged.
          tight_layout: Whether to apply tight layout to the figure before logging it.
          close: Whether to close the figure after logging it.
        """
        ...

    @abstractmethod
    def log_graph(self, graph: torch.nn.Module, data: TensorOrTensors, epoch: int) -> Self:
        """Log the given computation graph by tracing it with the given data.

        Alternatively, loggers may choose to log the graph using TorchScript, run
        it on the given data, or use any other mechanism they see fit.

        Parameters:
          graph: The computation graph to log, represented as a PyTorch module.
          data: The input data to use for tracing the computation graph.
          epoch: The epoch number at which the computation graph is logged.
        """
        ...

    @abstractmethod
    def log_image(self, label: str, image: AnyArray, epoch: int, data_format: DataFormat = "HWC") -> Self:
        """Log the given image with the given label at the given epoch.

        Parameters:
          label: The label of the logged image.
          image: The image to log, represented as a `PIL` image or as an array
            of any of the following shapes (assuming `"HWC"` data format):

            - `(H, W)` or `(H, W, 1)` for grayscale images,
            - `(H, W, 2)` for grayscale images with alpha channel,
            - `(H, W, 3)` for RGB images,
            - `(H, W, 4)` for RGBA images.

            If the pixel values are floating-point numbers, they are expected
            to be in the `[0, 1]` range; otherwise, they are assumed to be in the
            `[0, 255]` range.
          epoch: The epoch number at which the image is logged.
          data_format: The data format of the image specifying whether the channels
            are stored in the last dimension (`"HWC"`, the default) or in the first dimension (`"CHW"`);
            ignored for a `PIL` image.
        """
        ...

    @abstractmethod
    def log_metrics(self, logs: dict[str, float], epoch: int, description: str | None = None) -> Self:
        """Log metrics collected during a given epoch, with an optional description.

        Parameters:
          logs: A dictionary of logged metrics for the epoch.
          epoch: The epoch number at which the logs were collected.
          description: An optional description of the logged metrics (used only by some loggers).
        """
        ...

    @abstractmethod
    def log_text(self, label: str, text: str, epoch: int) -> Self:
        """Log the given text with the given label at the given epoch.

        Parameters:
          label: The label of the logged text.
          text: The text to log.
          epoch: The epoch number at which the text is logged.
        """
        ...



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Literal, TYPE_CHECKING

from ..callback import Callback
from ..type_aliases import Logs
if TYPE_CHECKING:
    from ..trainable_module import TrainableModule


class KeepBestWeights(Callback):
    """A callback that keeps the best model weights in memory."""

    def __init__(
        self,
        metric: str,
        mode: Literal["max", "min"] = "max",
        device: str | None = None,
    ) -> None:
        """Create the KeepBestWeights callback.

        Parameters:
          metric: The metric name from `logs` dictionary to monitor.
          mode: One of `"max"` or `"min"`, indicating whether the monitored metric should be maximized
            or minimized.
          device: The device where the weights will be stored. If `None`, the weights will be stored
            on the same device as the model.
        """
        assert mode in ("max", "min"), "mode must be one of 'max' or 'min'"

        self._metric = metric
        self._mode = mode
        self._device = device

        self.best_state_dict = None
        self.best_value = None

    best_state_dict: dict | None
    """The state dictionary containing the copies of best weights encountered so far."""

    best_value: float | None = None
    """The best metric value seen so far."""

    def __call__(self, module: "TrainableModule", epoch: int, logs: Logs) -> None:
        if (self.best_value is None
                or (self._mode == "max" and logs[self._metric] > self.best_value)
                or (self._mode == "min" and logs[self._metric] < self.best_value)):
            self.best_value = logs[self._metric]
            self.best_state_dict = {k: v.to(device=self._device, copy=True)
                                    for k, v in module.state_dict().items()}



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""The Keras-inspired high-level API for training PyTorch models.

The [TrainableModule][npfl138.TrainableModule] class is
a high-level API for training PyTorch models. It is a subclass of
[torch.nn.Module][] and:

- It provides a high-level API for training, evaluation, and prediction
  via [fit][npfl138.TrainableModule.fit], [evaluate][npfl138.TrainableModule.evaluate],
  and [predict][npfl138.TrainableModule.predict] methods. Each can be customized
  by overriding the corresponding [train_step][npfl138.TrainableModule.train_step],
  [test_step][npfl138.TrainableModule.test_step], or [predict_step][npfl138.TrainableModule.predict_step]
  methods.

- The module automatically handles moving the module to a specified device,
  using the first available accelerator by default. To this end,
  [configure][npfl138.TrainableModule.configure] or
  [load_weights][npfl138.TrainableModule.load_weights] must always be called
  first before using the high-level API.

- The module inputs and outputs can be either single tensors or _tensor structures_
  (the [TensorOrTensors][npfl138.type_aliases.TensorOrTensors] type), where the latter
  can be tuples, lists, or dictionaries containing other tensor structures and non-tensor
  values, or completely custom data structures. Tensors in tuples, lists, and dictionary
  values are automatically moved to the correct device during [fit][npfl138.TrainableModule.fit],
  [evaluate][npfl138.TrainableModule.evaluate], and [predict][npfl138.TrainableModule.predict].

- The module provides API for serialization and deserialization of the model,
  both the weights ([save_weights][npfl138.TrainableModule.save_weights],
  [load_weights][npfl138.TrainableModule.load_weights]) and the model options
  ([save_options][npfl138.TrainableModule.save_options],
  [load_options][npfl138.TrainableModule.load_options]).

- The module keeps a collection of tracked [losses][npfl138.TrainableModule.losses], and
  a collection of [metrics][npfl138.TrainableModule.metrics] implementing the
  [Metric][npfl138.Metric] interface (e.g., any our metric or any metric from `torchmetric` package),
  and is capable of storing the computed logs in text files, TensorBoard logs, WandB, or any
  implementation of the [Logger][npfl138.Logger] interface.
"""
import argparse
from collections.abc import Iterable
import functools
import json
import os
import sys
import time
from typing import Any, Literal, Self

import torch

from .callback import Callback, STOP_TRAINING
from .logger import Logger
from .loggers import BaseLogger, FileSystemLogger, MultiLogger, TensorBoardLogger
from .loss import Loss
from .metric import Metric
from .metrics import Mean
from .progress_logger import ProgressLogger as BaseProgressLogger
from .type_aliases import Logs, TensorOrTensors
from .utils import compute_logs, fill_and_standardize_path, tuple_list


class KeepPrevious:
    pass
keep_previous = KeepPrevious()  # noqa: E305


class ProgressLogger(BaseProgressLogger):
    @staticmethod
    def log_config(config: dict[str, Any], epoch: int, console: int | None, logger: Logger | None) -> None:
        if ProgressLogger.get_console_verbosity(console) >= 1:
            print(BaseLogger.format_config_as_text(config, epoch), flush=True)
        logger and logger.log_config(config, epoch)

    def log_epoch(self, logs: dict[str, float], epoch: int, elapsed: float, logger: Logger | None = None) -> None:
        description = self._description + f" {elapsed:.1f}s"
        self._console and print(description, BaseLogger.format_metrics(logs), flush=True)
        logger and logger.log_metrics(logs, epoch, description)


def get_auto_device() -> torch.device:
    """Return an available accelerator or CPU if none is available, unless overridden by `NPFL_DEVICE`."""
    if "NPFL_DEVICE" in os.environ:
        return torch.device(os.environ["NPFL_DEVICE"])
    if torch.cuda.is_available():
        return torch.device("cuda")
    if torch.backends.mps.is_available():
        return torch.device("mps")
    if torch.xpu.is_available():
        return torch.device("xpu")
    return torch.device("cpu")


class static_or_instance_method:
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, owner=None):
        return functools.partial(self.func, instance)


def tensors_concatenate(x: list[TensorOrTensors] | tuple[TensorOrTensors, ...]) -> TensorOrTensors:
    """Concatenate a list of tensors or tensor structures along the first dimension."""
    if not x:
        raise RuntimeError("Cannot concatenate an empty list of tensors.")
    first = x[0]
    if isinstance(first, torch.Tensor):
        return torch.cat(x)
    elif isinstance(first, torch.nn.utils.rnn.PackedSequence):
        raise RuntimeError("Concatenation of torch.nn.utils.rnn.PackedSequence is not supported.")
    elif isinstance(first, tuple):
        return tuple(map(tensors_concatenate, zip(*x)))
    elif isinstance(first, list):
        return [tensors_concatenate(b) for b in zip(*x)]
    elif isinstance(first, dict):
        return {k: tensors_concatenate([b[k] for b in x]) for k in first.keys()}
    raise RuntimeError(f"Cannot concatenate tensors of type {type(first)}.")


def tensors_to_device(x: TensorOrTensors, device: torch.device) -> TensorOrTensors:
    """Asynchronously move the input tensor or the input tensor structure to the given device.

    Parameters:
      x: The input tensor or tensor structure to move to the device, where tensor structures
        can be tuples, lists, or dictionaries containing other tensor structures and non-tensor
        values, or completely custom data structures. All tensors in tuples, lists, and dictionary
        values are moved.
      device: The device to move the tensors to.

    Returns:
      The input tensor or tensor structure with all tensors moved to the given device.
    """
    if isinstance(x, (torch.Tensor, torch.nn.utils.rnn.PackedSequence)):
        return x.to(device, non_blocking=True)
    elif isinstance(x, tuple):
        return tuple(tensors_to_device(a, device) for a in x)
    elif isinstance(x, list):
        return [tensors_to_device(a, device) for a in x]
    elif isinstance(x, dict):
        return {k: tensors_to_device(v, device) for k, v in x.items()}
    return x


def tensors_to_device_as_tuple(x: TensorOrTensors, device: torch.device) -> tuple[TensorOrTensors, ...]:
    """Asynchronously move input tensor data structure to a given device, returning a (possibly one-element) tuple."""
    if isinstance(x, (torch.Tensor, torch.nn.utils.rnn.PackedSequence)):
        return (x.to(device),)
    elif isinstance(x, tuple_list):
        return tuple(tensors_to_device(a, device) for a in x)
    elif isinstance(x, dict):
        return ({k: tensors_to_device(v, device) for k, v in x.items()},)
    else:
        return (x,)


def tensors_to_numpy(x: TensorOrTensors) -> TensorOrTensors:
    """Convert tensors in a given input tensor data structure to NumPy arrays."""
    if isinstance(x, torch.Tensor):
        return x.numpy(force=True)
    elif isinstance(x, torch.nn.utils.rnn.PackedSequence):
        raise RuntimeError("Conversion of torch.nn.utils.rnn.PackedSequence to Numpy is not supported.")
    elif isinstance(x, tuple):
        return tuple(tensors_to_numpy(b) for b in x)
    elif isinstance(x, list):
        return [tensors_to_numpy(b) for b in x]
    elif isinstance(x, dict):
        return {k: tensors_to_numpy(v) for k, v in x.items()}
    return x


def validate_batch_input_output(
    batch: tuple[TensorOrTensors, TensorOrTensors] | list[TensorOrTensors],
) -> tuple[TensorOrTensors, TensorOrTensors] | list[TensorOrTensors]:
    """Validate that the given batch is an (input, output) pair and return it."""
    assert isinstance(batch, tuple_list) and len(batch) == 2, "The batch must be an (input, output) pair."
    return batch


def validate_batch_input(
    batch: TensorOrTensors | tuple[TensorOrTensors, TensorOrTensors] | list[TensorOrTensors], with_labels: bool,
) -> TensorOrTensors:
    """If with_labels is True, validate that the given batch is an (input, _) pair and return the input."""
    if with_labels:
        assert isinstance(batch, tuple_list) and len(batch) == 2, "The batch must be an (input, _) pair."
        batch = batch[0]
    return batch


class TrainableModule(torch.nn.Module):
    """A simple Keras-like module for training with raw PyTorch.

    The module provides fit/evaluate/predict methods, computes loss and metrics,
    and generates TensorBoard, text file, and console logs. By default, it uses
    an accelerator (GPU, MPS, XPU) if available, and CPU otherwise.

    The input to the model can be either a single tensor/PackedSequence or
    a tuple of those. Similarly, the output can be a single tensor/PackedSequence
    or a tuple of those. However, when there are multiple outputs, you
    must handle loss and metrics computation manually.
    """

    def __init__(self, module: torch.nn.Module | None = None):
        """Initialize the module, optionally with an existing PyTorch module.

        Parameters:
          module: An optional existing PyTorch module to wrap, e.g., a [torch.nn.Sequential][]
            or a pretrained Transformer. If given, the module still must be configured.
        """
        super().__init__()
        setattr(self, "device", None)  # Avoid mkdocs documenting the attribute before its definition.
        setattr(self, "module", module)  # Avoid mkdocs documenting the attribute before its definition.
        self.unconfigure()
        if module is not None:
            self.forward = self._wrapped_module_forward

    def _wrapped_module_forward(self, inputs):
        return self.module(inputs)

    def configure(
        self,
        *,
        optimizer: torch.optim.Optimizer | None | KeepPrevious = keep_previous,
        scheduler: torch.optim.lr_scheduler.LRScheduler | None | KeepPrevious = keep_previous,
        loss: Loss | None | KeepPrevious = keep_previous,
        metrics: dict[str, Metric] | KeepPrevious = keep_previous,
        initial_epoch: int | KeepPrevious = keep_previous,
        logdir: str | None | KeepPrevious = keep_previous,
        loggers: list[Logger] | KeepPrevious = keep_previous,
        device: torch.device | str | Literal["auto"] | KeepPrevious = keep_previous,
    ) -> Self:
        """Configure the module fitting, evaluation, and placement.

        The method can be called multiple times, preserving previously set values by default.

        Note:
          The [self.device][npfl138.TrainableModule.device], [self.epoch][npfl138.TrainableModule.epoch],
          [self.logger][npfl138.TrainableModule.logger] and [self.metrics][npfl138.TrainableModule.metrics]
          fields are never `None` after this call (they are the only options for which `None` is not allowed).

        Parameters:
          optimizer: The optimizer to use for training.
          scheduler: An optional learning rate scheduler used after every batch.
          loss: The loss function to minimize, implementing the
            [Loss][npfl138.Loss] protocol.
          metrics: A dictionary of additional metrics to compute, each being an
            object implementing the metric (reset/update/compute) interface, e.g.,
            any [Metric][npfl138.Metric] or any `torchmetrics.Metric`.
          initial_epoch: The initial epoch of the model used during training and evaluation.
          logdir: An optional directory where textual logs should be stored; when specified,
            [npfl138.loggers.FileSystemLogger][] is added to the loggers.
          loggers: An optional list of loggers to use for logging; when not specified and
            `logdir` is given, [npfl138.loggers.TensorBoardLogger][] is added to the loggers
            (specifying a list of loggers including `[]` prevents that).
          device: The device to move the module to. When `"auto"`, or `keep_previous`
            with no previously set device, an accelerator [torch.accelerator.current_accelerator][]
            is used if available, otherwise a CPU; furthermore, environment variable `NPFL_DEVICE`
            can be used to override the "auto" device selection.

        Returns:
          self
        """
        self.optimizer = optimizer if optimizer is not keep_previous else self.optimizer
        self.scheduler = scheduler if scheduler is not keep_previous else self.scheduler
        self.loss = loss if loss is not keep_previous else self.loss
        if metrics is not keep_previous or self.metrics is None:
            self.metrics = torch.nn.ModuleDict({} if metrics is keep_previous else metrics)
        self.epoch = initial_epoch if initial_epoch is not keep_previous else self.epoch or 0
        if (logdir is not keep_previous and logdir != self.logdir) or loggers is not keep_previous:
            self.logger = self.logger and self.logger.close()
        if (logdir is not keep_previous and logdir != self.logdir) and logdir:
            loggers = [TensorBoardLogger(logdir)] if loggers is keep_previous else loggers
            loggers = [FileSystemLogger(logdir)] + loggers
        loggers = [] if self.logger is None and loggers is keep_previous else loggers
        self.logger = MultiLogger(loggers) if loggers is not keep_previous else self.logger
        self.logdir = logdir if logdir is not keep_previous else self.logdir
        if device is not keep_previous or self.device is None:
            self.device = get_auto_device() if device == "auto" or device is keep_previous else torch.device(device)
        return self.to(self.device)

    def unconfigure(self) -> Self:
        """Remove all training configuration of the TrainableModule.

        Only the module device is kept.

        Returns:
          self
        """
        getattr(self, "logger", None) is not None and self.logger.close()
        self.loss, self.optimizer, self.scheduler, self.epoch = None, None, None, None
        self.logdir, self.logger, self.losses, self.metrics = None, None, None, None
        return self

    def fit(
        self,
        dataloader: torch.utils.data.DataLoader,
        epochs: int,
        *,
        dev: torch.utils.data.DataLoader | None = None,
        callbacks: list[Callback] = [],
        log_config: dict[str, Any] = {},
        log_graph: bool = False,
        console: int | None = None,
    ) -> dict[str, float]:
        """Train the model on the given dataset.

        Note:
          To customize the training, you can override the [train_step][npfl138.TrainableModule.train_step],
          [compute_loss][npfl138.TrainableModule.compute_loss], and/or
          [compute_metrics][npfl138.TrainableModule.compute_metrics] method.

        Parameters:
          dataloader: The training dataset, each element a pair of inputs and outputs;
            the inputs and outputs can be either single tensors or tensor structures.
          epochs: The number of epochs to train.
          dev: An optional development dataset to evaluate after every epoch, with the
            same format as the training dataset.
          callbacks: A list of callbacks to call after every epoch, each implementing
            the [npfl138.Callback][] protocol with arguments `self`, `epoch`, and `logs`,
            possibly returning [npfl138.STOP_TRAINING][] to stop the training (note that
            the module is set to evaluation mode before calling each callback).
          log_config: Optionally log the given configuration dictionary at the beginning of training.
          log_graph: Optionally log the model graph at the beginning of training.
          console: Controls the console verbosity: 0 for silent, 1 for epoch logs, 2 for
            additional only-when-writing-to-console progress bar, 3 for persistent progress bar.
            The default is 2, but can be overridden by the `NPFL_PROGRESS` environment variable.

        Returns:
          logs: A dictionary of logs from the training and optionally dev evaluation.

        Note:
          The module is set to evaluation mode when returning from this method.
        """
        assert self.metrics is not None, "The TrainableModule has not been configured, run configure first."
        log_config and ProgressLogger.log_config(log_config, self.epoch, console, self.logger)
        logs, epochs, stop_training = {}, self.epoch + epochs, False
        while self.epoch < epochs and not stop_training:
            self.epoch += 1
            self.train()
            self.losses = torch.nn.ModuleDict()
            for metric in self.metrics.values():
                metric.reset()
            start, logs = time.time(), {}
            data_with_progress = ProgressLogger(dataloader, f"Epoch {self.epoch}/{epochs}", console, lambda: logs)
            for batch in data_with_progress:
                xs, y = validate_batch_input_output(batch)
                xs = tensors_to_device_as_tuple(xs, self.device)
                y = tensors_to_device(y, self.device)
                logs = self.train_step(xs, y)
                log_graph = log_graph and self.logger.log_graph(self.module or self, xs, self.epoch - 1) and False
            if dev is not None:
                compute_logs(logs)
                logs |= self.eval().evaluate(dev, "dev", log_results=False, console=console)
            for callback in callbacks:
                stop_training = callback(self.eval(), self.epoch, compute_logs(logs)) is STOP_TRAINING or stop_training
            data_with_progress.log_epoch(compute_logs(logs), self.epoch, time.time() - start, self.logger)
        self.eval()
        return logs

    def train_step(self, xs: TensorOrTensors, y: TensorOrTensors) -> Logs:
        """An overridable method performing a single training step, returning the logs.

        Parameters:
          xs: The input batch to the model, either a single tensor or a tensor structure.
          y: The target output batch of the model, either a single tensor or a tensor structure.

        Returns:
          logs: A dictionary of logs from the training step.
        """
        y_pred = self(*xs)
        loss = self.track_loss(self.compute_loss(y_pred, y, *xs))
        loss.backward()
        with torch.no_grad():
            self.optimizer.step()
            self.optimizer.zero_grad()
            self.scheduler is not None and self.scheduler.step()
            metrics = self.compute_metrics(y_pred, y, *xs)
            return {**({"lr": self.scheduler.get_last_lr()[0]} if self.scheduler else {}), **self.losses, **metrics}

    def compute_loss(
        self, y_pred: TensorOrTensors, y: TensorOrTensors, *xs: TensorOrTensors,
    ) -> torch.Tensor | dict[str, torch.Tensor]:
        """Compute the loss of the model given the inputs, predictions, and target outputs.

        Parameters:
          y_pred: The model predictions, either a single tensor or a tensor structure.
          y: The target output of the model, either a single tensor or a tensor structure.
          *xs: The inputs to the model; unpacked, if the input is a list or a tuple.

        Returns:
          loss: The computed loss, either a single tensor or a dictionary of named loss components.
            In case of a dictionary, the total loss is either the item called `"loss"` if present,
            or the sum of all items otherwise.
        """
        return self.loss(y_pred, y)

    def compute_metrics(self, y_pred: TensorOrTensors, y: TensorOrTensors, *xs: TensorOrTensors) -> Logs:
        """Compute and return metrics given the inputs, predictions, and target outputs.

        Parameters:
          y_pred: The model predictions, either a single tensor or a tensor structure.
          y: The target output of the model, either a single tensor or a tensor structure.
          *xs: The inputs to the model; unpacked, if the input is a list or a tuple.

        Returns:
          logs: A dictionary of computed metrics.
        """
        for metric in self.metrics.values():
            metric.update(y_pred, y)
        return self.metrics

    def track_loss(self, loss: torch.Tensor | dict[str, torch.Tensor]) -> torch.Tensor:
        """A method for manually tracking a loss value during training or evaluation.

        Parameters:
          loss: The loss value to track, either a single tensor or a dictionary of named loss components.
            In case of a dictionary, the total loss is either the item called `"loss"` if present,
            or the sum of all items otherwise.

        Returns:
          total_loss: The total loss tensor.
        """
        if isinstance(loss, dict):
            if "loss" not in loss:
                loss = {"loss": sum(loss.values())} | loss
        else:
            loss = {"loss": loss}

        for name, value in loss.items():
            if name not in self.losses:
                self.losses[name] = Mean(device=self.device)
            self.losses[name].update(value)

        return loss["loss"]

    def evaluate(
        self,
        dataloader: torch.utils.data.DataLoader,
        dataset_name: str | None = "test",
        *,
        log_results: bool = True,
        callbacks: list[Callback] = [],
        console: int | None = None,
    ) -> dict[str, float]:
        """An evaluation of the model on the given dataset.

        Note:
          To customize the evaluation, you can override the [test_step][npfl138.TrainableModule.test_step],
          [compute_loss][npfl138.TrainableModule.compute_loss], and/or
          [compute_metrics][npfl138.TrainableModule.compute_metrics] method.

        Parameters:
          dataloader: The dataset to evaluate on, each element a pair of inputs and outputs;
            the inputs and outputs can be either single tensors or tensor structures.
          dataset_name: An optional name of the dataset used as a prefix of the metric names in the logs.
          log_results: If `True` (the default), the evaluation results are logged using the module logger,
            and they are also printed (if `console` is not 0); if `False`, they are just returned.
          callbacks: A list of callbacks to call after the evaluation, each implementing
            the [npfl138.Callback][] protocol with arguments `self`, `epoch`, and `logs`.
          console: Controls the console verbosity: 0 for silent, 1 for a single message,
            2 for additional only-when-writing-to-console progress bar, 3 for persistent progress bar.
            The default is 2, but can be overridden by the `NPFL_PROGRESS` environment variable.

        Returns:
          logs: A dictionary of logs from the evaluation, each name prefixed with `f"{dataset_name}:"`
            if `dataset_name` is given.

        Note:
          The module is set to evaluation mode when returning from this method.
        """
        assert self.metrics is not None, "The TrainableModule has not been configured, run configure first."
        self.eval()
        self.losses = torch.nn.ModuleDict()
        for metric in self.metrics.values():
            metric.reset()
        start, logs = time.time(), {}
        data_with_progress = ProgressLogger(dataloader, "Evaluation", console, lambda: logs)
        for batch in data_with_progress:
            xs, y = validate_batch_input_output(batch)
            xs = tensors_to_device_as_tuple(xs, self.device)
            y = tensors_to_device(y, self.device)
            logs = self.test_step(xs, y)
            logs = {f"{dataset_name}:{k}": v for k, v in logs.items()} if dataset_name else logs
        for callback in callbacks:
            callback(self.eval(), self.epoch, compute_logs(logs))
        compute_logs(logs)
        log_results and data_with_progress.log_epoch(logs, self.epoch, time.time() - start, self.logger)
        self.eval()
        return logs

    def test_step(self, xs: TensorOrTensors, y: TensorOrTensors) -> Logs:
        """An overridable method performing a single evaluation step, returning the logs.

        Parameters:
          xs: The input batch to the model, either a single tensor or a tensor structure.
          y: The target output batch of the model, either a single tensor or a tensor structure.

        Returns:
          logs: A dictionary of logs from the evaluation step.
        """
        with torch.no_grad():
            y_pred = self(*xs)
            self.track_loss(self.compute_loss(y_pred, y, *xs))
            metrics = self.compute_metrics(y_pred, y, *xs)
            return {**self.losses, **metrics}

    def predict(
        self,
        dataloader: torch.utils.data.DataLoader,
        *,
        data_with_labels: bool = False,
        whole_batches: bool = False,
        as_numpy: bool = False,
        console: int | None = None,
    ) -> Iterable[TensorOrTensors]:
        """Compute predictions for the given dataset as a generator of individual items or whole batches.

        None:
          To customize the prediction, you can override the [predict_step][npfl138.TrainableModule.predict_step]
          and/or [unpack_batch][npfl138.TrainableModule.unpack_batch] methods.

        Parameters:
          dataloader: The dataset to predict on, each element either directly the module
            input or a pair whose first element is the input; the input can be either
            a single tensor or a tensor structure.
          data_with_labels: Specifies whether the dataloader elements are _(input, labels)_ pairs
            or just _inputs_ (the default).
          whole_batches: If `False` (the default), the predicted batches are unpacked into individual items
            using the [unpack_batch][npfl138.TrainableModule.unpack_batch] method; if `True`, whole predicted
            batches are yielded.
          as_numpy: If `False` (the default), the predicted tensors are kept as PyTorch tensors on the module device;
            if `True`, they are converted to Numpy arrays.
          console: Controls the console verbosity: 0 and 1 for silent, 2 for
            additional only-when-writing-to-console progress bar, 3 for persistent progress bar.
            The default is 2, but can be overridden by the `NPFL_PROGRESS` environment variable.

        Returns:
          predictions: An iterable whose elements are the individual predicted items.
        """
        assert self.device is not None, "No device has been set for the TrainableModule, run configure first."
        self.eval()
        for batch in ProgressLogger(dataloader, "Prediction", console):
            xs = validate_batch_input(batch, with_labels=data_with_labels)
            xs = tensors_to_device_as_tuple(xs, self.device)
            y = self.predict_step(xs)
            y = self.unpack_batch(y, *xs) if not whole_batches else [y]
            yield from map(tensors_to_numpy, y) if as_numpy else y

    def predict_step(self, xs: TensorOrTensors) -> TensorOrTensors:
        """An overridable method performing a single prediction step.

        Parameters:
          xs: The input batch to the model, either a single tensor or a tensor structure.

        Returns:
          predictions: The predicted batch.
        """
        with torch.no_grad():
            return self(*xs)

    def unpack_batch(self, y: TensorOrTensors, *xs: TensorOrTensors) -> Iterable[TensorOrTensors]:
        """An overridable method unpacking a batch of predictions into individual items.

        The default implementation handles batches of single [torch.Tensor][]s,
        [torch.nn.utils.rnn.PackedSequence][]s, or tuples, lists, and dictionaries
        containing elements (values in the case of dictionaries) that are themselves
        unpackable by this method.

        Parameters:
          y: The batch predictions, either a single tensor or a tensor structure.
          *xs: The inputs to the model; unpacked, if the input is a list or a tuple.

        Returns:
          items: An iterable over the individual predicted items.
        """
        if isinstance(y, torch.Tensor):
            yield from y
        elif isinstance(y, torch.nn.utils.rnn.PackedSequence):
            yield from torch.nn.utils.rnn.unpack_sequence(y)
        elif isinstance(y, tuple):
            yield from zip(*(self.predicted_batch_as_items(b) for b in y))
        elif isinstance(y, list):
            yield from map(list, zip(*(self.predicted_batch_as_items(b) for b in y)))
        elif isinstance(y, dict):
            for items in zip(*(self.predicted_batch_as_items(v) for v in y.values())):
                yield dict(zip(y.keys(), items))
        else:
            raise RuntimeError(f"Cannot unpack batch of type {type(y)} into individual items.")

    def predict_batch(self, xs: TensorOrTensors, *, as_numpy: bool = False) -> TensorOrTensors:
        """Run prediction on a single batch, returning the predicted batch.

        This method is a convenience wrapper around [predict_step][npfl138.TrainableModule.predict_step].
        It sets the module to evaluation mode, move the input to the module device, calls
        [predict_step][npfl138.TrainableModule.predict_step], and optionally converts the output to Numpy arrays.

        None:
          To customize prediction, you can override the [predict_step][npfl138.TrainableModule.predict_step] method.

        Warning:
          To avoid calling [torch.nn.Module.eval][] too often, `predict_batch` calls [torch.nn.Module.eval][]
          only if `self.training` is `True`.

        Parameters:
          xs: The input batch to the model, either a single tensor or a tensor structure. Note that it
            cannot be a pair of inputs and outputs.
          as_numpy: If `False` (the default), the predicted tensors are kept as PyTorch tensors on the module device;
            if `True`, they are converted to Numpy arrays.

        Returns:
          predictions: The predicted batch.
        """
        assert self.device is not None, "No device has been set for the TrainableModule, run configure first."
        self.training and self.eval()
        xs = tensors_to_device_as_tuple(xs, self.device)
        y = self.predict_step(xs)
        y = tensors_to_numpy(y) if as_numpy else y
        return y

    def predict_tensor(
        self,
        dataloader: torch.utils.data.DataLoader,
        *,
        data_with_labels: bool = False,
        as_numpy: bool = False,
        console: int | None = None,
    ) -> TensorOrTensors:
        """Compute predictions for the given dataset, returning them concatenated as a single tensor/tensor structure.

        This method is a convenience wrapper around [predict][npfl138.TrainableModule.predict].

        None:
          To customize prediction, you can override the [predict_step][npfl138.TrainableModule.predict_step] method.

        Parameters:
          dataloader: The dataset to predict on, each element either directly the module
            input or a pair whose first element is the input; the input can be either
            a single tensor or a tensor structure.
          data_with_labels: Specifies whether the dataloader elements are _(input, labels)_ pairs
            or just _inputs_ (the default).
          as_numpy: If `False` (the default), the predicted tensors are kept as PyTorch tensors on the module device;
            if `True`, they are converted to Numpy arrays.
          console: Controls the console verbosity: 0 and 1 for silent, 2 for
            additional only-when-writing-to-console progress bar, 3 for persistent progress bar.
            The default is 2, but can be overridden by the `NPFL_PROGRESS` environment variable.

        Returns:
          predictions: The predicted dataset concatenated to a single tensor or a tensor structure.
        """
        y = list(self.predict(dataloader, data_with_labels=data_with_labels, whole_batches=True, console=console))
        y = tensors_concatenate(y)
        return tensors_to_numpy(y) if as_numpy else y

    def save_weights(self, path: str, optimizer_path: str | None = None) -> Self:
        """Save the model weights to the given path.

        Both the `path` and `optimizer_path` can contain `{logdir}` and `{epoch}` placeholders.
        They are always processed with `str.format` before use, and both slashes and backslashes
        are replaced with the current OS path separator.

        Parameters:
          path: The path to save the model weights to; a `.pt` extension is recommended.
          optimizer_path: An optional path to save the optimizer state to, relative to the
            model weights path.

        Returns:
          self
        """
        path = fill_and_standardize_path(path, logdir=self.logdir, epoch=self.epoch)
        if optimizer_path is not None:
            optimizer_path = fill_and_standardize_path(optimizer_path, logdir=self.logdir, epoch=self.epoch)
        os.path.dirname(path) and os.makedirs(os.path.dirname(path), exist_ok=True)

        state_dict = self.state_dict()
        torch.save(state_dict, path)

        # Save the number of epochs, optimizer state, and the scheduler state when requested.
        if optimizer_path is not None:
            optimizer_state = {"epoch": self.epoch}
            self.optimizer is not None and optimizer_state.update(optimizer=self.optimizer.state_dict())
            self.scheduler is not None and optimizer_state.update(scheduler=self.scheduler.state_dict())
            optimizer_path = os.path.join(os.path.dirname(path), optimizer_path)
            os.path.dirname(optimizer_path) and os.makedirs(os.path.dirname(optimizer_path), exist_ok=True)
            torch.save(optimizer_state, optimizer_path)
        return self

    def load_weights(self, path: str, optimizer_path: str | None = None,
                     device: torch.device | str | Literal["auto"] | KeepPrevious = keep_previous) -> Self:
        """Load the model weights from the given path.

        Both the `path` and `optimizer_path` can contain `{logdir}` and `{epoch}` placeholders.
        They are always processed with `str.format` before use, and both slashes and backslashes
        are replaced with the current OS path separator.

        Parameters:
          path: The path to load the model weights from.
          optimizer_path: An optional path to load the optimizer state from, relative to the
            model weights path.
          device: The device to load the module to. When `"auto"`, or `keep_previous`
            with no previously set device, an accelerator [torch.accelerator.current_accelerator][]
            is used if available, otherwise a CPU; furthermore, environment variable `NPFL_DEVICE`
            can be used to override the "auto" device selection.

        Returns:
          self
        """
        path = fill_and_standardize_path(path, logdir=self.logdir, epoch=self.epoch)
        if optimizer_path is not None:
            optimizer_path = fill_and_standardize_path(optimizer_path, logdir=self.logdir, epoch=self.epoch)

        if device is not keep_previous or not self.device:
            self.device = get_auto_device() if device == "auto" or device is keep_previous else torch.device(device)
        self.load_state_dict(torch.load(path, map_location=self.device))

        # Load the number of epochs, optimizer state, and the scheduler state when requested.
        if optimizer_path is not None:
            optimizer_path = os.path.join(os.path.dirname(path), optimizer_path)
            optimizer_state = torch.load(optimizer_path, map_location=self.device)
            self.epoch = optimizer_state["epoch"]
            if self.optimizer is not None:
                assert "optimizer" in optimizer_state, "The optimizer state is missing."
                self.optimizer.load_state_dict(optimizer_state["optimizer"])
            else:
                assert "optimizer" not in optimizer_state, "The optimizer state is present, but there is no optimizer."
            if self.scheduler is not None:
                assert "scheduler" in optimizer_state, "The scheduler state is missing."
                self.scheduler.load_state_dict(optimizer_state["scheduler"])
            else:
                assert "scheduler" not in optimizer_state, "The scheduler state is present, but there is no scheduler."
        return self.to(self.device)

    @static_or_instance_method
    def save_options(self, _path_: str, /, **kwargs: Any) -> None:
        """Save a JSON-serializable options or [argparse.Namespace][]s to the given path.

        The method can be called both as a static method and as an instance method.

        When called as an instance method, the path can contain `{logdir}` and `{epoch}` placeholders.
        The path is always processed with `str.format` before use, and both slashes
        and backslashes are replaced with the current OS path separator.

        Parameters:
          _path_: The path to save the options to; a `.json` extension is recommended. The parameter
            name uses the underscores to avoid conflict with possible option `path`.
          **kwargs: The option values to save.
        """
        path = fill_and_standardize_path(_path_, **{"logdir": self.logdir, "epoch": self.epoch} if self else {})
        options = dict((k + ": argparse.Namespace", vars(v)) if isinstance(v, argparse.Namespace) else (k, v)
                       for k, v in kwargs.items())
        os.path.dirname(path) and os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, "w", encoding="utf-8") as options_file:
            json.dump(options, options_file, ensure_ascii=False, indent=2)

    @staticmethod
    def load_options(path: str) -> dict:
        """Load a JSON-serializable options or [argparse.Namespace][]s from the given path.

        Parameters:
          path: The path to load the options from.

        Returns:
          options: The loaded options dictionary.
        """
        with open(path, "r", encoding="utf-8-sig") as options_file:
            options = json.load(options_file)
        return dict((k.removesuffix(": argparse.Namespace"), argparse.Namespace(**v))
                    if k.endswith(": argparse.Namespace") else (k, v) for k, v in options.items())

    def profile(
        self,
        steps: int,
        export_trace: str | None = None,
        *,
        warmup: int | None = 0,
        lightweight: bool = False,
        export_memory_timeline: str | bool = False,
        export_cuda_allocations: str | bool = False,
        quit_when_done: bool = False,
    ) -> None:
        """Profile the module execution for a number of steps.

        Run the PyTorch profiler on a CPU and an accelerator if available (and optionally track CUDA allocations),
        for the given number of steps (forward passes) after an optional number of warmup steps.

        By default, the profiler records detailed information including shapes, stack traces, and memory utilization,
        which has additional overhead. When `lightweight=True`, only basic information is recorded.

        Info:
          The exported profile trace can be inspected in TensorBoard using the `torch-tb-profiler` plugin that
          can be installed using `pip install torch-tb-profiler`.

        Info:
          The CUDA memory allocations snapshot can be inspected using the <https://docs.pytorch.org/memory_viz> page.

        Parameters:
          steps: The number of steps (forward calls) to profile. For example, when `steps=2`, the profiler starts
            at the beginning of the first step (forward) and stops at the beginning of the third step (forward).
          export_trace: An optional path to export the main profile to (as a Chrome trace JSON file). The file must
            have an extension of either `.pt.trace.json` or `.pt.trace.json.gz` to be recognized by the
            `torch-tb-profiler` plugin; if the path does not end with one of those extensions, `.pt.trace.json.gz`
            is appended.
          warmup: An optional number of warmup steps to skip before starting the profiling.

            - When 0 (the default), profiling starts at the beginning of the first step (forward call).
            - When 1, profiling starts at the beginning of the second step (forward call).
            - When `None`, the profiling starts immediately (which can be useful to track CUDA allocations
              during module initialization).
          lightweight: If `True`, use a lightweight profiling mode that records only basic information, excluding
            tensor shapes, stack traces, and memory utilization. This reduces the profiling overhead.
          export_memory_timeline: An optional path to export the memory timeline HTML report to. If a string is
            passed, it is used as the path (appending `.html` if needed); if `True` is passed, the path is derived
            from `export_trace` by replacing the extension with `.html`.

            **Note**: Requires the `matplotlib` package for generating the graph.
          export_cuda_allocations: An optional path to export the CUDA memory allocations snapshot to (when CUDA
            is available). If a string is passed, it is used as the path (appending `.pickle` if needed);
            if `True` is passed, the path is derived from `export_trace` by replacing the extension with `.pickle`.
          quit_when_done: If `True`, the program exits when profiling is done.
        """
        if lightweight and export_memory_timeline is not False:
            raise ValueError("The export_memory_timeline argument cannot be used with lightweight=True.")

        # Standardize all export paths.
        if export_trace is not None:
            if not export_trace.endswith(".pt.trace.json") and not export_trace.endswith(".pt.trace.json.gz"):
                export_trace += ".pt.trace.json.gz"
            export_trace = fill_and_standardize_path(export_trace, logdir=self.logdir)

        if isinstance(export_memory_timeline, str):
            if not export_memory_timeline.endswith(".html"):
                export_memory_timeline += ".html"
            export_memory_timeline = fill_and_standardize_path(export_memory_timeline, logdir=self.logdir)
        elif export_memory_timeline is True:
            assert export_trace is not None, "export_trace must be specified when export_memory_timeline is True."
            export_memory_timeline = export_trace.rsplit(".pt.trace.json", maxsplit=1)[0] + ".html"

        if isinstance(export_cuda_allocations, str):
            if not export_cuda_allocations.endswith(".pickle"):
                export_cuda_allocations += ".pickle"
            export_cuda_allocations = fill_and_standardize_path(export_cuda_allocations, logdir=self.logdir)
        elif export_cuda_allocations is True:
            assert export_trace is not None, "export_trace must be specified when export_cuda_allocations is True."
            export_cuda_allocations = export_trace.rsplit(".pt.trace.json", maxsplit=1)[0] + ".pickle"

        # Set up the profiler hook.
        profiler, hook = None, None

        def profile_step(_module, _inputs) -> None:
            nonlocal steps, warmup, profiler
            if warmup > 0:
                warmup -= 1
            elif steps > 0:
                if profiler is None:
                    if export_cuda_allocations and torch.cuda.is_available():
                        torch.cuda.memory._record_memory_history()
                    # We use `acc_events=True` to avoid false-positive warning in PyTorch 2.10 about acc_events.
                    profiler = torch.profiler.profile(
                        profile_memory=not lightweight, record_shapes=not lightweight, with_stack=not lightweight,
                        acc_events=True)
                    profiler.__enter__()
                steps -= 1
            elif profiler is not None:
                profiler.__exit__(None, None, None)
                if export_cuda_allocations and torch.cuda.is_available():
                    torch.cuda.memory._dump_snapshot(export_cuda_allocations)
                    torch.cuda.memory._record_memory_history(enabled=None)
                if export_trace:
                    profiler.export_chrome_trace(export_trace)
                if export_memory_timeline:
                    profiler.export_memory_timeline(export_memory_timeline)
                hook.remove()
                profiler = None
                quit_when_done and sys.exit(0)

        # Register the profiler hook.
        hook = self.register_forward_pre_hook(profile_step)

        # When warmup is `None`, start profiling immediately.
        if warmup is None:
            warmup = 0
            steps += 1
            profile_step(None, None)

    def log_console(
        self, message: str, end: str = "\n", progress_only: bool = False, console: int | None = None,
    ) -> Self:
        """Write the given message to the console, correctly even if a progress bar is being used.

        Parameters:
          message: The message to write.
          end: The string appended after the message.
          progress_only: If `False` (the default), the message is written to standard output when current console
            verbosity is at least 1; if `True`, the message is written to standard error only when the progress bar
            is being shown (console verbosity 2 and writing to the console, or console verbosity 3).
          console: Controls the current console verbosity. The default is 2, but can be overridden by the
            `NPFL_PROGRESS` environment variable.
        """
        ProgressLogger.log_console(message, end, progress_only, console)
        return self

    def get_tb_writer(self, name: str) -> torch.utils.tensorboard.SummaryWriter:
        """Possibly create and return a TensorBoard writer for the given name.

        To use this method, a [npfl138.loggers.TensorBoardLogger][] must have been
        created in the [configure][npfl138.TrainableModule.configure] method, either by
        specifying a `logdir` and no `loggers` or by explicitly passing the TensorBoard logger.

        Returns:
          writer: The opened TensorBoard writer.
        """
        for logger in self.logger.loggers:
            if isinstance(logger, TensorBoardLogger):
                return logger.get_writter(name)
        raise RuntimeError("No TensorBoardLogger found in the configured loggers.")

    device: torch.device | None
    """The device where the module is located, if already set."""

    epoch: int | None
    """The current one-based epoch of the module, if already set."""

    logdir: str | None
    """The directory where the logs are stored, if set."""

    loss: Loss | None
    """The loss function used for training, if set."""

    logger: MultiLogger | None
    """The logger used for logging the training and evaluation, if set."""

    losses: torch.nn.ModuleDict | None
    """The dictionary of tracked losses used in training and evaluation, if set."""

    metrics: torch.nn.ModuleDict | None
    """The dictionary of metrics used for training and evaluation, if set."""

    module: torch.nn.Module | None
    """The wrapped PyTorch module, if specified during construction."""

    optimizer: torch.optim.Optimizer | None
    """The optimizer used for training, if set."""

    scheduler: torch.optim.lr_scheduler.LRScheduler | None
    """The learning rate scheduler used for training, if set."""



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Self

import torch

from .mean import Mean
from ..utils import broadcast_to_prefix


class CategoricalAccuracy(Mean):
    """Categorical classification accuracy metric.

    The predictions are assumed to be logits or probabilities predicted by a model,
    while the ground-truth targets are either class indices (sparse format) or whole
    distributions (dense format). In both cases, the predicted class is considered
    to be the one with the highest probability.
    """

    def __init__(
        self, dim: int = 1, *, ignore_index: int = -100, probs: bool = False, device: torch.device | None = None,
    ) -> None:
        """Create the CategoricalAccuracy metric object.

        Parameters:
          dim: If the input has 2 or more dimensions, this value specifies the dimension along which
            the classes are defined. The default is the same behavior as [torch.nn.CrossEntropyLoss][].
          ignore_index: An optional target class value that is ignored during metric computation (equivalent
            to zeroing out sample weights for the corresponding samples). The default of -100 is the same
            as in [torch.nn.CrossEntropyLoss][].
          probs: For consistency with other categorical losses and metrics, we include this parameter
            describing whether the predictions are logits or probabilities. However, to predict the
            most probable class, it does not matter whether logits or probabilities are used.
        """
        super().__init__(device)
        self._dim = dim
        self._ignore_index = ignore_index

    @torch.no_grad
    def update(
        self, y: torch.Tensor, y_true: torch.Tensor, sample_weights: torch.Tensor | None = None,
    ) -> Self:
        """Update the accumulated categorical accuracy using new predictions and gold labels.

        Optional sample weight might be provided; if not, all values are weighted with 1.

        Parameters:
          y: The predicted outputs, either logits or probabilities (depending on the `probs` parameter).
            If they have 2 or more dimensions, the class dimension is specified by the `dim` parameter.
            We consider the class with the highest probability to be predicted.
          y_true: The ground-truth targets in two possible formats:

            - The gold targets might be "sparse" class indices. In this case, their shape has to be
              exactly the same as `y` with the class dimension removed.
            - The gold targets might be full "dense" probability distributions. In this case, their
              shape has to be exactly the same as `y` and we consider the class with the highest probability
              to be the gold class.
          sample_weights: Optional sample weights. If provided, their shape must be broadcastable
            to a prefix of a shape of `y` with the class dimension removed, and the loss for each sample
            is weighted accordingly.

        Returns:
          self
        """
        y_shape, y_true_shape = y.shape, y_true.shape
        dim = self._dim % len(y_shape) if len(y_shape) > 1 else 0
        y_wo_class_dim_shape = y_shape[:dim] + y_shape[dim + 1:]

        dense = len(y_true_shape) == len(y_shape)
        if dense:
            assert y_true_shape == y_shape, "In dense format, y_true must have the same shape as y."
            y_true = torch.argmax(y_true, dim=dim)
        else:
            assert y_true_shape == y_wo_class_dim_shape, \
                "In sparse format, y_true must have the same shape as y with the class dimension removed."
            y_true_dtype = y_true.dtype
            assert not y_true_dtype.is_floating_point and not y_true_dtype.is_complex, \
                "In sparse format, y_true must contain class indices."
            if y_true_dtype != torch.int64 and y_true_dtype != torch.uint8:
                y_true = y_true.long()

        y = torch.argmax(y, dim=dim)

        ignore_index_weights = (y_true != self._ignore_index).to(dtype=torch.float32)
        if sample_weights is None:
            sample_weights = ignore_index_weights
        else:
            sample_weights = broadcast_to_prefix(sample_weights, y_wo_class_dim_shape)
            sample_weights = sample_weights.to(dtype=torch.float32) * ignore_index_weights

        return super().update(y == y_true, sample_weights=sample_weights)



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Self

import torch

from .mean import Mean
from .. import losses


class BinaryCrossEntropy(Mean):
    """Binary cross-entropy metric implementation."""

    def __init__(
        self, *, label_smoothing: float = 0.0, probs: bool = False, device: torch.device | None = None,
    ) -> None:
        """Create the BinaryCrossEntropy metric object.

        Parameters:
          label_smoothing: A float in [0.0, 1.0] specifying the label smoothing factor.
            If greater than 0.0, the used ground-truth targets are computed as a mixture
            of the original targets and uniform distribution with weight `1 - label_smoothing`.
          probs: If `False`, the predictions are assumed to be logits; if `True`, the
            predictions are assumed to be probabilities. Note that gold targets are
            always expected to be probabilities.
        """
        super().__init__(device)
        self._bce_loss = losses.BinaryCrossEntropy(label_smoothing=label_smoothing, probs=probs, reduction="none")

    @torch.no_grad
    def update(
        self, y: torch.Tensor, y_true: torch.Tensor, sample_weights: torch.Tensor | None = None,
    ) -> Self:
        """Update the accumulated binary cross-entropy by introducing new values.

        Optional sample weight might be provided; if not, all values are weighted with 1.

        Parameters:
          y: The predicted outputs. Their shape either has to be exactly the same as `y_true` (no broadcasting),
            or can contain an additional single dimension of size 1.
          y_true: The ground-truth targets.
          sample_weights: Optional sample weights. If provided, their shape must be broadcastable
            to a prefix of a shape of `y_true`, and the loss for each sample is weighted accordingly.

        Returns:
          self
        """
        return super().update(self._bce_loss(y, y_true), sample_weights=sample_weights)



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Self

import torch

from .mean import Mean
from .. import losses
from ..utils import broadcast_to_prefix


class CategoricalCrossEntropy(Mean):
    """Categorical cross-entropy metric implementation."""

    def __init__(
        self,
        dim: int = 1,
        *,
        ignore_index: int = -100,
        label_smoothing: float = 0.0,
        probs: bool = False,
        device: torch.device | None = None,
    ) -> None:
        """Create the CategoricalCrossEntropy metric object.

        Parameters:
          dim: If the input has 2 or more dimensions, this value specifies the dimension along which
            the classes are defined. The default is the same behavior as [torch.nn.CrossEntropyLoss][].
          ignore_index: An optional target class value that is ignored during loss computation (equivalent
            to zeroing out sample weights for the corresponding samples). Only applicable for sparse targets;
            when dense targets are used, the default of -100 cannot be overwritten and this parameter is
            ignored. This is the same behavior as [torch.nn.CrossEntropyLoss][].
          label_smoothing: A float in [0.0, 1.0] specifying the label smoothing factor.
            If greater than 0.0, the used ground-truth targets are computed as a mixture
            of the original targets and uniform distribution with weight `1 - label_smoothing`.
          probs: If `False`, the predictions are assumed to be logits; if `True`, the
            predictions are assumed to be probabilities. Note that gold targets are
            always expected to be probabilities in the dense format.
        """
        super().__init__(device)
        self._cce_loss = losses.CategoricalCrossEntropy(
            dim=dim, ignore_index=ignore_index, label_smoothing=label_smoothing, probs=probs, reduction="none",
        )
        self._dim = dim
        self._ignore_index = ignore_index

    @torch.no_grad
    def update(
        self, y: torch.Tensor, y_true: torch.Tensor, sample_weights: torch.Tensor | None = None,
    ) -> Self:
        """Update the accumulated categorical cross-entropy by introducing new values.

        Optional sample weight might be provided; if not, all values are weighted with 1.

        Parameters:
          y: The predicted outputs, either logits or probabilities (depending on the `probs` parameter).
            If they have 2 or more dimensions, the class dimension is specified by the `dim` parameter.
          y_true: The ground-truth targets in two possible formats:

            - The gold targets might be "sparse" class indices. In this case, their shape has to be
              exactly the same as `y` with the class dimension removed.
            - The gold targets might be full "dense" probability distributions. In this case, their
              shape has to be exactly the same as `y`.
          sample_weights: Optional sample weights. If provided, their shape must be broadcastable
            to a prefix of a shape of `y` with the class dimension removed, and the loss for each sample
            is weighted accordingly.

        Returns:
          self
        """
        y_shape, y_true_shape = y.shape, y_true.shape

        dense = len(y_true_shape) == len(y_shape)
        if not dense:
            y_true_dtype = y_true.dtype
            assert not y_true_dtype.is_floating_point and not y_true_dtype.is_complex, \
                "In sparse format, y_true must contain class indices."
            if y_true_dtype != torch.int64 and y_true_dtype != torch.uint8:
                y_true = y_true.long()

            # For sparse targets, we need to mask out the ignored classes manually.
            ignore_index_weights = (y_true != self._ignore_index).to(dtype=torch.float32)
            if sample_weights is None:
                sample_weights = ignore_index_weights
            else:
                dim = self._dim % len(y_shape) if len(y_shape) > 1 else 0
                y_wo_class_dim_shape = y_shape[:dim] + y_shape[dim + 1:]
                sample_weights = broadcast_to_prefix(sample_weights, y_wo_class_dim_shape)
                sample_weights = sample_weights.to(dtype=torch.float32) * ignore_index_weights

        return super().update(self._cce_loss(y, y_true), sample_weights=sample_weights)



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from collections.abc import Iterable


class Vocabulary:
    """A class for managing mapping between strings and indices.

    The vocabulary is initialized with a list of strings, and additionally can contain
    two special tokens:

    - a padding token [npfl138.Vocabulary.PAD_TOKEN][], which, if present, is always at index
      [npfl138.Vocabulary.PAD][]=0;
    - an unknown token [npfl138.Vocabulary.UNK_TOKEN][], which, if present, is either at index
      [npfl138.Vocabulary.UNK][] 0 or 1 (depending on whether the padding token is present);
      the index of this token is returned when looking up a string not present in the vocabulary.

    Info:
      A `Vocabulary` instance can be pickled and unpickled efficiently as a list of strings;
      the required string-to-index mapping is reconstructed upon unpickling.
    """
    PAD: int | None
    """The index of the padding token, either `None` or `0`."""
    PAD_TOKEN: str = "[PAD]"
    """The string representing the padding token."""

    UNK: int | None
    """The index of the unknown token, either `None`, `0`, or `1`."""
    UNK_TOKEN: str = "[UNK]"
    """The string representing the unknown token."""

    def __init__(self, strings: Iterable[str], add_pad: bool = False, add_unk: bool = False) -> None:
        """Initialize the vocabulary with the given list of strings.

        The strings might be prepended with special tokens for padding and unknown tokens, respectively,
        depending on the values of `add_pad` and `add_unk`.

        Note:
          If the given strings already contain special tokens on expected indices, they are recognized
          correctly and no duplicates are added even if `add_pad` and/or `add_unk` are `True`.

        Parameters:
          strings: An iterable of strings to include in the vocabulary.
          add_pad: Whether to add a padding token [npfl138.Vocabulary.PAD_TOKEN][] at index 0
            and set [npfl138.Vocabulary.PAD][]=0.
          add_unk: Whether to add an unknown token [npfl138.Vocabulary.UNK_TOKEN][] at index 0 or 1
            (depending on whether the padding token is added) and set [npfl138.Vocabulary.UNK][] accordingly.
        """
        # Get the first two strings (if any) to check for special tokens.
        it, head = iter(strings), []
        try:
            head.append(next(it))
            head.append(next(it))
        except StopIteration:
            it = None

        # Start by adding the special tokens.
        self._strings = []

        if add_pad or (head and head[0] == self.PAD_TOKEN):  # Add PAD token if required or present.
            self.PAD = 0
            self._strings.append(self.PAD_TOKEN)
            (head and head[0] == self.PAD_TOKEN) and head.pop(0)
        else:
            self.PAD = None

        if add_unk or (head and head[0] == self.UNK_TOKEN):  # Add UNK token if required or present.
            self.UNK = len(self._strings)
            (head and head[0] == self.UNK_TOKEN) or self._strings.append(self.UNK_TOKEN)
        else:
            self.UNK = None

        # Now add the remaining strings, both from `head` and from `it`.
        self._strings.extend(head)
        it is not None and self._strings.extend(it)

        self._string_map = {string: index for index, string in enumerate(self._strings)}

    def __len__(self) -> int:
        """The number of strings in the vocabulary.

        Returns:
          The size of the vocabulary.
        """
        return len(self._strings)

    def __iter__(self) -> Iterable[str]:
        """Return an iterator over strings in the vocabulary.

        Returns:
          An iterator over strings in the vocabulary.
        """
        return iter(self._strings)

    def __getstate__(self) -> dict:
        state = self.__dict__.copy()
        del state["_string_map"]
        return state

    def __setstate__(self, state: dict) -> None:
        self.__dict__.update(state)
        self._string_map = {string: index for index, string in enumerate(self._strings)}

    def add(self, string: str) -> int:
        """If not already present, add the given string to the end of the vocabulary.

        Parameters:
          string: The string to add.

        Returns:
          The index of the newly added string (or the index of the existing string if it was already present).
        """
        index = self._string_map.get(string)
        if index is None:
            index = len(self._strings)
            self._strings.append(string)
            self._string_map[string] = index
        return index

    def string(self, index: int) -> str:
        """Convert vocabulary index to string.

        Parameters:
          index: The vocabulary index.

        Returns:
          The string corresponding to the given index.
        """
        return self._strings[index]

    def strings(self, indices: Iterable[int]) -> list[str]:
        """Convert a sequence of vocabulary indices to strings.

        Parameters:
          indices: An iterable of vocabulary indices.

        Returns:
          A list of strings corresponding to the given indices.
        """
        return [self._strings[index] for index in indices]

    def index(self, string: str, add_missing: bool = False) -> int | None:
        """Convert string to vocabulary index.

        Parameters:
          string: The string to convert.
          add_missing: Whether to add the string to the vocabulary if not present.

        Returns:
          The index corresponding to the given string. If the string is not found in the vocabulary, then

            - if `add_missing` is `True`, the string is added to the end of the vocabulary and its index returned;
            - if the [npfl138.Vocabulary.UNK_TOKEN][] was added to the vocabulary, its index is returned;
            - otherwise, `None` is returned.
        """
        if add_missing:
            return self.add(string)
        else:
            return self._string_map.get(string, self.UNK)

    def indices(self, strings: Iterable[str], add_missing: bool = False) -> list[int | None]:
        """Convert a sequence of strings to vocabulary indices.

        Parameters:
          strings: An iterable of strings to convert.
            add_missing: Whether to add strings not present in the vocabulary.

        Returns:
          A list of indices corresponding to the given strings. For each string not found in the vocabulary:

            - if `add_missing` is `True`, the string is added to the end of the vocabulary and its index returned;
            - if the [npfl138.Vocabulary.UNK_TOKEN][] was added to the vocabulary, its index is returned;
            - otherwise, `None` is returned.
        """
        if add_missing:
            return [self.add(string) for string in strings]
        else:
            return [self._string_map.get(string, self.UNK) for string in strings]



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import torch

from ..loss import Loss
from ..type_aliases import Reduction
from ..utils import broadcast_to_prefix

cross_entropy_loss = torch.nn.functional.cross_entropy


class CategoricalCrossEntropy(Loss):
    """Categorical cross-entropy loss implementation."""

    def __init__(
        self,
        dim: int = 1,
        *,
        ignore_index: int = -100,
        label_smoothing: float = 0.0,
        probs: bool = False,
        reduction: Reduction = "mean",
    ) -> None:
        """Create the CategoricalCrossEntropy loss object with the specified reduction method.

        Parameters:
          dim: If the input has 2 or more dimensions, this value specifies the dimension along which
            the classes are defined. The default is the same behavior as [torch.nn.CrossEntropyLoss][].
          ignore_index: An optional target class value that is ignored during loss computation (equivalent
            to zeroing out sample weights for the corresponding samples). Only applicable for sparse targets;
            when dense targets are used, the default of -100 cannot be overwritten and this parameter is
            ignored. This is the same behavior as [torch.nn.CrossEntropyLoss][].
          label_smoothing: A float in [0.0, 1.0] specifying the label smoothing factor.
            If greater than 0.0, the used ground-truth targets are computed as a mixture
            of the original targets and uniform distribution with weight `1 - label_smoothing`.
          probs: If `False`, the predictions are assumed to be logits; if `True`, the
            predictions are assumed to be probabilities. Note that gold targets are
            always expected to be probabilities in the dense format.
          reduction: The reduction method to apply to the computed loss.
        """

        self._dim = dim
        self._ignore_index = ignore_index
        self._label_smoothing = label_smoothing
        self._probs = probs
        self._reduction = reduction

    def __call__(
        self, y: torch.Tensor, y_true: torch.Tensor, sample_weights: torch.Tensor | None = None,
    ) -> torch.Tensor:
        """Compute the categorical cross-entropy loss, optionally with sample weights.

        Parameters:
          y: The predicted outputs, either logits or probabilities (depending on the `probs` parameter).
            If they have 2 or more dimensions, the class dimension is specified by the `dim` parameter.
          y_true: The ground-truth targets in two possible formats:

            - The gold targets might be "sparse" class indices. In this case, their shape has to be
              exactly the same as `y` with the class dimension removed.
            - The gold targets might be full "dense" probability distributions. In this case, their
              shape has to be exactly the same as `y`.
          sample_weights: Optional sample weights. If provided, their shape must be broadcastable
            to a prefix of a shape of `y` with the class dimension removed, and the loss for each sample
            is weighted accordingly.

        Returns:
          A tensor representing the computed loss. A scalar tensor if reduction is `"mean"` or `"sum"`;
            otherwise (if reduction is `"none"`), a tensor of the same shape as `y` without the class dimension.
        """
        y_shape, y_true_shape = y.shape, y_true.shape
        dim = self._dim % len(y_shape) if len(y_shape) > 1 else 0
        y_wo_class_dim_shape = y_shape[:dim] + y_shape[dim + 1:]

        dense = len(y_true_shape) == len(y_shape)
        if dense:
            assert y_true_shape == y_shape, "In dense format, y_true must have the same shape as y."
            assert self._ignore_index == -100, "When ignore_index is set, y_true cannot be in dense format."
        else:
            assert y_true_shape == y_wo_class_dim_shape, \
                "In sparse format, y_true must have the same shape as y with the class dimension removed."
            y_true_dtype = y_true.dtype
            assert not y_true_dtype.is_floating_point and not y_true_dtype.is_complex, \
                "In sparse format, y_true must contain class indices."
            if y_true_dtype != torch.int64 and y_true_dtype != torch.uint8:
                y_true = y_true.long()

        if self._probs:
            y = y.clamp(min=1e-7, max=1.0 - 1e-7).log()

        if len(y_shape) > 1 and dim != 1:
            y = y.movedim(dim, 1)
            if dense:
                y_true = y_true.movedim(dim, 1)

        if sample_weights is None:
            return cross_entropy_loss(y, y_true, ignore_index=self._ignore_index,
                                      label_smoothing=self._label_smoothing, reduction=self._reduction)
        else:
            sample_weights = broadcast_to_prefix(sample_weights, y_wo_class_dim_shape)

            losses = sample_weights * cross_entropy_loss(y, y_true, ignore_index=self._ignore_index,
                                                         label_smoothing=self._label_smoothing, reduction="none")
            if self._reduction == "none":
                return losses
            elif self._reduction == "sum":
                return losses.sum()
            else:  # self._reduction == "mean"
                return losses.sum() / sample_weights.sum()



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from collections.abc import Iterable
from typing import Any, Self

import torch

from .base_logger import BaseLogger
from ..logger import Logger
from ..type_aliases import AnyArray, DataFormat, TensorOrTensors


class MultiLogger(BaseLogger):
    """A logger that forwards all logging calls to multiple other loggers."""

    def __init__(self, loggers: Iterable[Logger]) -> None:
        """Initialize the multi-logger.

        Parameters:
          loggers: An iterable of loggers to forward the logging calls to.
        """
        self.loggers: list[Logger] = list(loggers)

    loggers: list[Logger]
    """The list of loggers to forward the logging calls to."""

    def close(self) -> None:
        for logger in self.loggers:
            logger.close()

    def log_audio(self, label: str, audio: AnyArray, sample_rate: int, epoch: int) -> Self:
        for logger in self.loggers:
            logger.log_audio(label, audio, sample_rate, epoch)
        return self

    def log_config(self, config: dict[str, Any], epoch: int) -> Self:
        for logger in self.loggers:
            logger.log_config(config, epoch)
        return self

    def log_figure(self, label: str, figure: Any, epoch: int, tight_layout: bool = True, close: bool = True) -> Self:
        import matplotlib.pyplot as plt
        for i, logger in enumerate(self.loggers):
            logger.log_figure(label, figure, epoch, tight_layout, False)
        close and plt.close(figure)
        return self

    def log_graph(self, graph: torch.nn.Module, data: TensorOrTensors, epoch: int) -> Self:
        for logger in self.loggers:
            logger.log_graph(graph, data, epoch)
        return self

    def log_image(self, label: str, image: AnyArray, epoch: int, data_format: DataFormat = "HWC") -> Self:
        for logger in self.loggers:
            logger.log_image(label, image, epoch, data_format)
        return self

    def log_metrics(self, logs: dict[str, float], epoch: int, description: str | None = None) -> Self:
        for logger in self.loggers:
            logger.log_metrics(logs, epoch, description)
        return self

    def log_text(self, label: str, text: str, epoch: int) -> Self:
        for logger in self.loggers:
            logger.log_text(label, text, epoch)
        return self



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import datetime
import inspect
import os
import re
from typing import Any

from .utils import fill_and_standardize_path, sanitize_path


def format_logdir(logdir_template: str, **kwargs: Any) -> str:
    """Format the log directory path by filling in placeholders.

    The `logdir_template` is formatted using `str.format`, where the `{key}` placeholders
    are replaced by the corresponding values from `kwargs`. Importantly, several
    placeholders are always provided automatically:

    - `{config}`: A comma-separated list of `key=value` pairs for sorted key-value items in `kwargs`.
        The keys are abbreviated to their first character per segment (with segments separated by
        hyphens or underscores). The maximum length of the placeholder is limited to 200 characters;
        if exceeded, the longest entries are truncated with ellipses (`...`) to fit within the limit.
    - `{file}`: The base name of the script file (without extension) that called this function;
        empty string if called from an interactive environment (e.g., Jupyter notebook).
    - `{timestamp}`: The current date and time in the format `YYYYMMDD_HHMMSS`.

    Path-unsafe characters in the placeholder values are replaced with underscores, and for convenience,
    several additional variants of each placeholder are supported:

    - `{key-}`, `{key_}`: same as `{key}`, but with additional hyphen/underscore if the value is non-empty,
    - `{-key}`, `{_key}`: same as `{key}`, but with leading hyphen/underscore if the value is non-empty.

    Finally, both slashes and backslashes are replaced with the current OS path separator.

    Parameters:
      logdir_template: The log directory template with placeholders.
      **kwargs: The keyword arguments to fill the template.

    Returns:
      The formatted log directory path.

    Example:
      ```python
      parser = argparse.ArgumentParser()
      ...
      args = parser.parse_args()

      logdir = npfl138.format_logdir("logs/{file-}{timestamp}{-config}", **vars(args))
      ```
    """
    # Create {config} placeholder.
    items = [(re.sub("(.)[^-_]*[-_]?", r"\1", str(k)), str(v)) for k, v in sorted(kwargs.items())]
    if sum(len(k) + 1 + min(len(v), 5) + 1 for k, v in items) - 1 > 200:
        raise ValueError("Signature is too long to fit even with maximum truncation.")

    if items:
        limit = max(len(v) for k, v in items)
        while sum(len(k) + 1 + min(len(v), limit) + 1 for k, v in items) - 1 > 200:  # guaranteed False when limit == 5
            limit -= 1
        items = [(k, v if len(v) <= limit else v[:limit // 2 - 1] + "..." + v[-limit // 2 + 2:]) for k, v in items]
    kwargs["config"] = ",".join(f"{k}={v}" for k, v in items)

    # Create {file} placeholder.
    current_frame = inspect.currentframe()
    if current_frame and current_frame.f_back:
        kwargs["file"] = os.path.splitext(os.path.basename(current_frame.f_back.f_globals.get("__file__", "")))[0]
    else:
        kwargs["file"] = ""

    # Create {timestamp} placeholder.
    kwargs["timestamp"] = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

    # Sanitize placeholder values and create variants.
    for key in list(kwargs.keys()):
        value = sanitize_path(str(kwargs[key]))
        kwargs[key] = value
        for separator in ["-", "_"]:
            kwargs[f"{key}{separator}"] = value and f"{value}{separator}"
            kwargs[f"{separator}{key}"] = value and f"{separator}{value}"

    return fill_and_standardize_path(logdir_template, **kwargs)



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Literal, Protocol, TYPE_CHECKING

from .type_aliases import Logs
if TYPE_CHECKING:
    from .trainable_module import TrainableModule


class StopTraining:
    pass
STOP_TRAINING = StopTraining()  # noqa: E305
"""A sentinel value that can be returned by a callback to stop training."""


class Callback(Protocol):
    def __call__(self, module: "TrainableModule", epoch: int, logs: Logs) -> Literal[STOP_TRAINING] | None:
        """Represents a callback called after every training epoch.

        If the callback returns [npfl138.STOP_TRAINING][], the training stops.

        Parameters:
          module: the module being trained
          epoch: the current epoch number (starting from one)
          logs: a dictionary of current logs, with all values being evaluated to float;
            newly computed metrics can be added here

        **Returns:**

          - [`npfl138.STOP_TRAINING`][npfl138.STOP_TRAINING] to stop the training,
          - [`None`][None] to continue.
        """
        ...



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from collections.abc import Callable, Iterable
import os
import sys

import tqdm

from .loggers import BaseLogger
from .type_aliases import Logs
from .utils import compute_logs


class ProgressLogger(tqdm.tqdm):
    """A slim wrapper around `tqdm.tqdm` for showing a progress bar, optionally with logs."""

    monitor_interval = 0  # Disable internal monitoring thread.

    _report_only_first = int(os.environ.get("NPFL_PROGRESS_FIRST", -1))  # Optional global limit to first N reports.

    @staticmethod
    def get_console_verbosity(console: int | None) -> int:
        if console is None and "NPFL_PROGRESS" in os.environ:
            console = int(os.environ["NPFL_PROGRESS"])
        elif console is None and ("NPFL_PROGRESS_FIRST" in os.environ or "NPFL_PROGRESS_EACH" in os.environ):
            console = 3 if ProgressLogger._report_only_first != 0 else 1
        elif console is None:
            console = 2
        return console

    def __init__(
        self,
        data: Iterable,
        description: str,
        console: int | None = None,
        logs_fn: Callable[[], Logs] | None = None,
    ) -> None:
        """Create a ProgressLogger instance.

        Parameters:
          data: Any iterable data to wrap, usually a [torch.utils.data.DataLoader][].
          description: A description string to show in front of the progress bar.
          console: Controls the console verbosity: 0 and 1 for silent, 2 for
            only-when-writing-to-console progress bar, 3 for persistent progress bar.
            The default is 2, but can be overridden by the `NPFL_PROGRESS` environment variable.
          logs_fn: An optional function returning the current logs to show alongside the progress bar.
            If given, the logs are fully computed and shown on each refresh.
        """
        console = self.get_console_verbosity(console)

        kwargs = {}
        if "NPFL_PROGRESS_EACH" in os.environ:
            kwargs["miniters"] = int(os.environ["NPFL_PROGRESS_EACH"])
            kwargs["mininterval"] = None

        self._console = console
        self._description = description
        self._logs_fn = logs_fn
        super().__init__(data, unit="batch", leave=False, disable=None if console == 2 else console < 2, **kwargs)

    def refresh(self, nolock=False, lock_args=None):
        if ProgressLogger._report_only_first > 0:
            ProgressLogger._report_only_first -= 1
        elif ProgressLogger._report_only_first == 0:
            return

        description = self._description
        if self._logs_fn is not None:
            description += (description and " ") + BaseLogger.format_metrics(compute_logs(self._logs_fn()))
        self.set_description(description, refresh=False)

        super().refresh(nolock=nolock, lock_args=lock_args)

    @staticmethod
    def log_console(message: str, end: str = "\n", progress_only: bool = False, console: int | None = None) -> None:
        """Write the given message to the console, correctly even if a progress bar is being used.

        Parameters:
          message: The message to write.
          end: The string appended after the message.
          progress_only: If `False` (the default), the message is written to standard output when current console
            verbosity is at least 1; if `True`, the message is written to standard error only when the progress bar
            is being shown (console verbosity 2 and writing to the console, or console verbosity 3).
          console: Controls the current console verbosity. The default is 2, but can be overridden by the
            `NPFL_PROGRESS` environment variable.
        """
        console = ProgressLogger.get_console_verbosity(console)
        if progress_only and ((console == 2 and sys.stderr.isatty()) or console >= 3):
            tqdm.tqdm.write(message, end=end, file=sys.stderr)
        elif (not progress_only) and console >= 1:
            tqdm.tqdm.write(message, end=end, file=sys.stdout)



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from .binary_accuracy import BinaryAccuracy
from .binary_cross_entropy import BinaryCrossEntropy
from .categorical_accuracy import CategoricalAccuracy
from .categorical_cross_entropy import CategoricalCrossEntropy
from .exact_match import ExactMatch
from .mean import Mean
from .mean_squared_error import MeanSquaredError



# This file is part of NPFL138 <http://github.com/ufal/npfl138/>.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from collections.abc import Callable
import os
from typing import Any

import torch

from .utils import tuple_list


class TransformedDataset(torch.utils.data.Dataset):
    """A dataset capable of applying transformations to its items and batches.

    Assuming the `TransformedDataset` is used within a [DataLoader][torch.utils.data.DataLoader],
    batches are produced as follows:

    - First, every element is retrieved from the source dataset using `__getitem__` (or a list of
      all batch elements at the same time using `__getitems__`).
    - Then, if [transform][npfl138.TransformedDataset.transform] is defined, it is applied to each individual item.
    - Next, if [collate][npfl138.TransformedDataset.collate] is defined, it is applied to the list of items to form
      a batch.
    - Finally, if [transform_batch][npfl138.TransformedDataset.transform_batch] is defined, it is applied to the batch.

    Warning:
      Given how PyTorch [torch.utils.data.DataLoader][] works, when specifying
      [collate][npfl138.TransformedDataset.collate] or [transform_batch][npfl138.TransformedDataset.transform_batch],
      the `collate_fn` of the DataLoader **must** be set to `self.collate_fn`.

      This is automatically done when using the [dataloader][npfl138.TransformedDataset.dataloader]
      method of this class to create the [DataLoader][torch.utils.data.DataLoader]. However,
      if you create a [DataLoader][torch.utils.data.DataLoader] manually, you **must** pass
      `collate_fn=transformed_dataset.collate_fn` or otherwise [collate][npfl138.TransformedDataset.collate]
      and [transform_batch][npfl138.TransformedDataset.transform_batch] will be ignored.
    """
    def __init__(self, dataset: torch.utils.data.Dataset, dataset_limit: int | None = None) -> None:
        """Create a new transformed dataset using the provided dataset with an optional limit.

        Parameters:
          dataset: The source dataset implementing `__len__` and `__getitem__`.
          dataset_limit: If given, limits the length of the dataset to this value.

        **Environment variables:** The following environment variable can be used
        to override the method parameters:

        - `NPFL_DATASET_LIMIT`: If set to a positive integer, overrides the `dataset_limit` parameter.
        """
        self._dataset = dataset
        self._dataset_has_getitems = hasattr(self._dataset, "__getitems__")

        if os.environ.get("NPFL_DATASET_LIMIT", "").isdecimal():
            dataset_limit = int(os.environ["NPFL_DATASET_LIMIT"])
        if dataset_limit is not None and dataset_limit < len(self._dataset):
            self._dataset = torch.utils.data.Subset(self._dataset, list(range(dataset_limit)))

    def __len__(self) -> int:
        """Return the number of items in the dataset."""
        return len(self._dataset)

    def __getitem__(self, index: int) -> Any:
        """Return the item at the specified index."""
        item = self._dataset[index]
        if self.transform is not None:
            return self.transform(*item) if isinstance(item, tuple_list) else self.transform(item)
        return item

    def __getitems__(self, indices: list[int]) -> Any:
        """Return a batch of items at the specified indices."""
        if self._dataset_has_getitems:
            batch = self._dataset.__getitems__(indices)
        else:
            batch = [self._dataset[i] for i in indices]
        if self.transform is not None:
            batch = [self.transform(*item) if isinstance(item, tuple_list) else self.transform(item) for item in batch]
        return batch

    @property
    def dataset(self) -> torch.utils.data.Dataset:
        """Return the source dataset."""
        return self._dataset

    transform: Callable | None = None
    """If given, `transform` is called on each item before returning it.

    If the dataset item is a tuple or a list, `transform` is called with it unpacked.
    """

    collate: Callable | None = None
    """If given, `collate` is called on a list of items before returning them as a batch."""

    transform_batch: Callable | None = None
    """If given, `transform_batch` is called on a batch before returning it.

    If the batch is a tuple or a list, `transform_batch` is called with it unpacked.
    """

    def collate_fn(self, batch: list[Any]) -> Any:
        """A function for a DataLoader to collate a batch of items using `collate` and/or `transform_batch`.

        This function is used as the `collate_fn` parameter of a DataLoader when `collate` or `transform_batch` is set.

        Parameters:
          batch: A list of items to collate and/or pass through `transform_batch`.
        """
        batch = self.collate(batch) if self.collate is not None else torch.utils.data.dataloader.default_collate(batch)
        if self.transform_batch is not None:
            batch = self.transform_batch(*batch) if isinstance(batch, tuple_list) else self.transform_batch(batch)
        return batch

    def dataloader(
        self, batch_size=1, *, shuffle=False, seed=None, num_workers=0, **kwargs: Any,
    ) -> torch.utils.data.DataLoader:
        """Create a DataLoader for this dataset.

        This method is a convenience wrapper around [torch.utils.data.DataLoader][]
        setting up the required parameters. Most arguments are passed directly to the
        [torch.utils.data.DataLoader][], with a few exceptions:

        - When `seed` is given, it is used to construct the `generator` argument for the
          DataLoader using `torch.Generator().manual_seed(seed)`; the `generator` options
          must not be specified in `kwargs`.
        - When `shuffle` is `False` and no `generator` is given, `torch.Generator()` is passed
          as `generator`. Otherwise, the global random number generator would be used during
          every construction of an iterator, i.e. during every `iter(dataloader)` call.
        - When `num_workers` is greater than 0, `persistent_workers` is set to True.
        - When `collate` or `transform_batch` is set, the `self.collate_fn` is passed as the
          `collate_fn` parameter.
        """
        if seed is not None:
            # If a seed is given, create a generator with the seed and pass it to the DataLoader.
            if kwargs.get("generator", None) is not None:
                raise ValueError("When seed is given, generator must not be specified.")
            kwargs["generator"] = torch.Generator().manual_seed(seed)

        if not shuffle and kwargs.get("generator", None) is None:
            # If not shuffling and no generator is given, pass an explicit generator to the Dataloader.
            # Otherwise, the global random generator would generate a number on every iter(dataloader) call.
            kwargs["generator"] = torch.Generator()

        if num_workers > 0:
            # By default, set persistent_workers to True, but allow it to be overridden
            kwargs.setdefault("persistent_workers", True)

        if self.collate is not None or self.transform_batch is not None:
            if "collate_fn" in kwargs:
                raise ValueError("When collate or transform_batch is overridden, collate_fn must not be given.")
            kwargs["collate_fn"] = self.collate_fn

        # Create and return the DataLoader
        return torch.utils.data.DataLoader(
            self, batch_size=batch_size, shuffle=shuffle, num_workers=num_workers, **kwargs)
